from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from celery import shared_task
from slack_sdk.errors import SlackClientError

from firefighter.slack.slack_app import DefaultWebClient, slack_client

if TYPE_CHECKING:
    from slack_sdk.web.client import WebClient

logger = logging.getLogger(__name__)


@shared_task(
    name="slack.send_message",
    autoretry_for=(SlackClientError,),
    retry_kwargs={"max_retries": 2},
    default_retry_delay=30,
)
@slack_client
def send_message(  # pylint: disable=keyword-arg-before-vararg
    client: WebClient = DefaultWebClient, *args: Any, **kwargs: Any
) -> dict[str, Any] | bytes:
    """Sends a message to a channel. All arguments are passed to the Slack API."""
    return client.chat_postMessage(*args, **kwargs).data



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from firefighter.firefighter.utils import get_in
from firefighter.slack.slack_app import DefaultWebClient, slack_client

if TYPE_CHECKING:
    from slack_sdk.models.views import View
    from slack_sdk.web.client import WebClient

logger = logging.getLogger(__name__)


@slack_client
def open_modal(
    view: View,
    trigger_id: str | None = None,
    body: dict[str, Any] | None = None,
    client: WebClient = DefaultWebClient,
) -> None:
    if trigger_id is None and body is None:
        raise ValueError("Please provide either a trigger_id or a body containing one!")

    if trigger_id is None and body is not None:
        trigger_id = body.get("trigger_id")

    if trigger_id is None:
        raise ValueError("No trigger_id found!")
    client.views_open(trigger_id=trigger_id, view=view)


@slack_client
def push_modal(
    view: View,
    trigger_id: str | None = None,
    body: dict[str, Any] | None = None,
    client: WebClient = DefaultWebClient,
) -> None:
    if trigger_id is None and body is None:
        raise ValueError("Please provide either a trigger_id or a body containing one!")

    if trigger_id is None and body is not None:
        trigger_id = body.get("trigger_id")
    if trigger_id is None:
        raise ValueError("No trigger_id found!")
    client.views_push(trigger_id=trigger_id, view=view)


@slack_client
def update_modal(
    view: View,
    trigger_id: str | None = None,
    body: dict[str, Any] | None = None,
    view_id: str | None = None,
    external_id: str | None = None,
    hash_value: str | None = None,
    client: WebClient = DefaultWebClient,
    **kwargs: Any,
) -> None:
    if body is None and trigger_id is None and (view_id is None or external_id is None):
        raise ValueError("We need a body or args here!")

    if hash_value is None:
        hash_value = get_in(body, "view.hash")

    if trigger_id is None and body is not None:
        trigger_id = body.get("trigger_id")

    if external_id is None:
        external_id = get_in(body, "view.external_id")

    if view_id is None:
        view_id = get_in(body, "container.view_id")
        if view_id is None:
            view_id = get_in(body, "view.id")

    client.views_update(
        trigger_id=trigger_id,
        view_id=view_id,
        external_id=external_id,
        hash=hash_value,
        view=view,
        **kwargs,
    )



from __future__ import annotations

from logging import Filter, Formatter, LogRecord
from typing import Any

from firefighter.firefighter.settings.settings_utils import ENV, config
from firefighter.logging.custom_json_formatter import (
    CustomJsonEncoder,
    CustomJsonFormatter,
)
from firefighter.logging.pretty_formatter import PrettyFormatter


class AccessLogFilter(Filter):
    """Filter Gunicorn.access logs to avoid logging static files and healthcheck requests logging."""

    def filter(self, record: LogRecord) -> bool:
        if not record.args:
            return True

        raw_uri: str = record.args["{raw_uri}e"]  # type: ignore

        if (
            raw_uri == "/api/v2/firefighter/monitoring/healthcheck"
            and record.levelno <= 20
        ):
            return False
        if raw_uri.startswith("/static/") and record.levelno <= 20:
            return False
        if raw_uri == "/favicon.ico" and record.levelno <= 30:  # noqa: SIM103
            return False
        return True


def get_json_formatter() -> dict[str, type[Formatter] | Any]:
    attr_whitelist = {"name", "levelname", "pathname", "lineno", "funcName"}
    attrs = [x for x in CustomJsonFormatter.RESERVED_ATTRS if x not in attr_whitelist]
    return {
        "()": CustomJsonFormatter,
        "json_indent": None,
        "json_encoder": CustomJsonEncoder,
        "reserved_attrs": attrs,
        "timestamp": True,
    }


base_level = "DEBUG" if ENV == "dev" else "INFO"
base_level_override = config("LOG_LEVEL", cast=str, default="")
if base_level_override and base_level_override != "":
    base_level = base_level_override.upper()

formatter: dict[str, type[Formatter] | Any]
formatter = {"()": PrettyFormatter} if ENV == "dev" else get_json_formatter()

FF_LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "dynamicfmt": formatter,
    },
    "handlers": {
        "console": {"class": "logging.StreamHandler", "formatter": "dynamicfmt"},
    },
    "loggers": {
        "django": {
            "handlers": ["console"],
            "propagate": False,
        },
        "watchfiles.main": {"level": "INFO"},
        "django.utils.autoreload": {"level": "INFO"},
        "django.request": {"handlers": ["console"], "propagate": False},
        "django.server": {"handlers": ["console"], "propagate": False},
        "django.template": {"handlers": ["console"], "propagate": False},
        "django.db.backends": {
            "handlers": ["console"],
            "propagate": False,
            "level": base_level,
        },
        "django.db.backends.schema": {"handlers": ["console"], "propagate": False},
        "gunicorn.access": {
            "handlers": ["console"],
            "filters": ["accessfilter"],
            "propagate": False,
        },
        "gunicorn.error": {"handlers": ["console"], "propagate": False},
        "ddtrace": {
            "handlers": ["console"],
            "level": "WARNING",
        },
        "faker.factory": {
            "level": "INFO",
        },
        "fsevents": {
            "level": "INFO",
        },
    },
    "filters": {"accessfilter": {"()": AccessLogFilter}},
    "root": {
        "handlers": ["console"],
        "level": base_level,
        "propagate": False,
    },
}

LOGGING = FF_LOGGING



from __future__ import annotations

import logging
import uuid
from datetime import datetime
from typing import TYPE_CHECKING, Any

import django_filters
from django.contrib.postgres.search import SearchQuery, SearchRank, SearchVector
from django.db import models
from django.db.models import (
    Count,
    F,
    OuterRef,
    Q,
    QuerySet,
    Subquery,
    Sum,
    Value,
)
from django.db.models.fields import DurationField
from django.db.models.functions import Cast
from django.urls import reverse
from django.utils import timezone
from django_filters.filters import ModelMultipleChoiceFilter
from django_stubs_ext.db.models import TypedModelMeta

from firefighter.firefighter.fields_forms_widgets import (
    CustomCheckboxSelectMultiple,
    FFDateRangeSingleFilterEmpty,
)
from firefighter.incidents.models.group import Group
from firefighter.incidents.models.metric_type import IncidentMetric

if TYPE_CHECKING:
    from django.conf import settings


logger = logging.getLogger(__name__)
TZ = timezone.get_current_timezone()


class ComponentManager(models.Manager["Component"]):
    model: type[Component]

    def queryset_with_mtbf(
        self,
        date_from: datetime,
        date_to: datetime,
        queryset: QuerySet[Component] | None = None,
        metric_type: str = "time_to_fix",
        field_name: str = "mtbf",
    ) -> QuerySet[Component]:
        """Returns a queryset of components with an additional `mtbf` field."""
        date_to = min(date_to, datetime.now(tz=TZ))

        date_interval = date_to - date_from
        queryset = queryset or self.get_queryset()

        return (
            queryset.order_by("group__order", "order")
            .annotate(
                metric_subquery=Subquery(
                    IncidentMetric.objects.filter(
                        incident__component=OuterRef("pk"),
                        metric_type__type=metric_type,
                        incident__created_at__gte=date_from,
                        incident__created_at__lte=date_to,
                    )
                    .values("incident__component")
                    .annotate(sum_downtime=Sum("duration"))
                    .values("sum_downtime")
                )
            )
            .annotate(
                incident_count=Count(
                    "incident",
                    filter=Q(
                        incident__created_at__gte=date_from,
                    )
                    & Q(
                        incident__created_at__lte=date_to,
                    ),
                ),
                incidents_downtime=F("metric_subquery"),
                incident_uptime=Value(date_interval) - F("incidents_downtime"),
            )
            .annotate(**{
                field_name: Cast(
                    F("incident_uptime") / F("incident_count"),
                    output_field=DurationField(),
                )
            })
        )

    @staticmethod
    def search(
        queryset: QuerySet[Component] | None, search_term: str
    ) -> tuple[QuerySet[Component], bool]:
        # XXX Common search method
        if queryset is None:
            queryset = Component.objects.all()

        # If not search, return the original queryset
        if search_term is None or search_term.strip() == "":
            return queryset, False
        # Postgres search on title + description
        # XXX Improve search performance and relevance
        # XXX Support partial word search (infra matches infrastructure)
        vector = (
            SearchVector("name", config="english", weight="A")
            + SearchVector(
                "description",
                config="english",
                weight="B",
            )
            + SearchVector(
                "group__name",
                config="english",
                weight="C",
            )
            + SearchVector(
                "group__description",
                config="english",
                weight="D",
            )
        )
        query = SearchQuery(search_term, config="english", search_type="websearch")
        queryset = (
            queryset.annotate(rank=SearchRank(vector, query))
            .filter(rank__gte=0.01)
            .order_by("-rank")
        )

        return queryset, False


class Component(models.Model):
    objects: ComponentManager = ComponentManager()

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=128)
    description = models.TextField(blank=True)
    order = models.IntegerField(
        default=0,
        help_text="Order of the component in the list. Should be unique per `Group`.",
    )
    group = models.ForeignKey(Group, on_delete=models.PROTECT)
    private = models.BooleanField(
        default=False,
        help_text="If true, incident created with this component won't be communicated, and conversations will be made private. This is useful for sensitive components. In the future, private incidents may be visible only to its members.",
    )
    deploy_warning = models.BooleanField(
        default=True,
        help_text="If true, a warning will be sent when creating an incident of high severity with this component.",
    )

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    if TYPE_CHECKING:
        # ruff: noqa: PLC0415
        if settings.ENABLE_SLACK:
            from firefighter.slack.models.conversation import Conversation
            from firefighter.slack.models.user_group import UserGroup

            usergroups: QuerySet[UserGroup]
            conversations: QuerySet[Conversation]

    class Meta(TypedModelMeta):
        ordering = ["order"]

    def __str__(self) -> str:
        return f"{'🔒 ' if self.private else ''}{self.name}"

    def get_absolute_url(self) -> str:
        return reverse("incidents:component-detail", kwargs={"component_id": self.id})


class ComponentFilterSet(django_filters.FilterSet):
    """Set of filters for Component, share by Web UI and API."""

    id = django_filters.CharFilter(lookup_expr="iexact")
    private = django_filters.BooleanFilter()

    group = ModelMultipleChoiceFilter(
        queryset=Group.objects.order_by("order"),
        field_name="group_id",
        label="Group",
        widget=CustomCheckboxSelectMultiple,
    )
    metrics_period = FFDateRangeSingleFilterEmpty(
        method="metrics_period_filter",
        label="MTBF period",
    )
    search = django_filters.CharFilter(
        field_name="search", method="component_search", label="Search"
    )

    @staticmethod
    def component_search(
        queryset: QuerySet[Component], _name: str, value: str
    ) -> QuerySet[Component]:
        """Search incidents by title, description, and ID.

        Args:
            queryset (QuerySet[Component]): Queryset to search in.
            _name:
            value (str): Value to search for.

        Returns:
            QuerySet[Component]: Search results.
        """
        return Component.objects.search(queryset=queryset, search_term=value)[0]

    @staticmethod
    def metrics_period_filter(
        queryset: QuerySet[Component],
        _name: str,
        value: tuple[datetime, datetime, Any, Any],
    ) -> QuerySet[Component]:
        gte, lte, _, _ = value
        return Component.objects.queryset_with_mtbf(gte, lte, queryset=queryset)



from __future__ import annotations

import logging
from typing import Any

from django.views import generic

from firefighter.pagerduty.models import PagerDutyEscalationPolicy, PagerDutyOncall

logger = logging.getLogger(__name__)


class OncallListView(generic.ListView[PagerDutyOncall]):
    template_name = "pages/oncall_list.html"
    context_object_name = "oncalls"

    @staticmethod
    def get_queryset() -> list[tuple[PagerDutyEscalationPolicy, list[PagerDutyOncall]]]:  # type: ignore[override]
        return PagerDutyOncall.objects.get_current_oncalls_per_escalation_policy()

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """No *args to pass."""
        # Call the base implementation first to get a context
        context = super().get_context_data(**kwargs)
        last_fetched = (
            PagerDutyOncall.objects.values("updated_at").order_by("-updated_at").first()
        )
        context["last_updated"] = last_fetched["updated_at"] if last_fetched else None
        context["page_title"] = "On-call Overview"
        return context



from __future__ import annotations

import logging
from datetime import datetime, timedelta
from functools import reduce
from typing import TYPE_CHECKING, Any, NotRequired, TypedDict

from dateutil.relativedelta import relativedelta
from django.contrib import messages
from django.db.models import Count, F, OuterRef, Q, Subquery, Value
from django.db.models.fields import FloatField, IntegerField
from django.db.models.functions import Cast
from django.template.defaultfilters import floatformat
from django.utils import timezone

from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.models import Group, Incident, Priority
from firefighter.incidents.models.metric_type import IncidentMetric
from firefighter.incidents.views.date_filter import get_date_range_from_special_date
from firefighter.incidents.views.date_utils import get_biggest_date

if TYPE_CHECKING:
    from django.db.models.query import QuerySet
    from django.http.request import HttpRequest
    from django_stubs_ext import WithAnnotations
logger = logging.getLogger(__name__)


def weekly_dashboard_context(
    request: HttpRequest, incidents: QuerySet[Incident]
) -> dict[str, Any]:
    logger.debug(request)

    # Dates for our queries
    date_gte, date_lte, date_str, date_input = get_date_range_from_parameters(request)

    if date_input and not (date_gte or date_lte):
        messages.error(request, "Invalid date range.")

    incident_by_priority, incident_by_priority_total_row = get_incident_by_priority(
        incidents
    )

    incident_age_by_priority = get_incident_age_by_priority(
        incidents, date_lte, date_gte
    )
    (
        incident_ttr_by_priority_total_row,
        incident_ttr_by_priority,
    ) = get_incident_time_to_by_priority(incidents)

    incidents_by_domain = get_incidents_by_domain(incidents)

    incidents_by_priority_chart = {
        "keys": [x.name for x in incident_by_priority],
        "values": [x.incident_total for x in incident_by_priority],
    }
    incidents_by_domain_chart = {
        "keys": [x.name for x in incidents_by_domain],
        "values": [x.incidents_nb for x in incidents_by_domain],
    }
    incident_by_status_chart = get_incident_by_status_chart_data(incidents)

    incident_by_priority_table = {
        "cols": ["Priority", "Total", "Opened", "Closed", "Closed %"],
        "body": {
            "objects": incident_by_priority,
            "format": [
                {"filter": None, "suffix": None, "key": "name"},
                {"filter": floatformat, "suffix": None, "key": "incident_total"},
                {"filter": floatformat, "suffix": None, "key": "incident_open"},
                {"filter": floatformat, "suffix": None, "key": "incident_closed"},
                {
                    "filter": floatformat,
                    "filter_args": 2,
                    "suffix": "%",
                    "key": "incident_closed_percentage",
                },
            ],
        },
        "footer": incident_by_priority_total_row,
    }

    # TODO Improve table template to allow cell customization (link)
    incident_age_by_priority_table = {
        "cols": [
            "Priority",
            "0-2 days",
            "3-4 days",
            "5-6 days",
            "7-14 days",
            "15-29 days",
            "30-89 days",
            "90+ days",
        ],
        "body": {
            "objects": incident_age_by_priority,
            "format": [
                {"key": "name"},
                {"key": "incident_0_3_label"},
                {"key": "incident_3_5_label"},
                {"key": "incident_5_7_label"},
                {"key": "incident_7_15_label"},
                {"key": "incident_15_30_label"},  # gitleaks:allow
                {"key": "incident_30_90_label"},
                {"key": "incident_gt_90_label"},
            ],
        },
    }

    incident_ttr_by_priority_percentage_table = {
        "cols": ["Priority", "< 15m", "< 30m", "< 1h", "< 3h", "< 1d", "≥ 1d"],
        "body": {
            "objects": incident_ttr_by_priority,
            "format": [
                {"filter": None, "suffix": None, "key": "name"},
                {
                    "filter": floatformat,
                    "filter_args": 1,
                    "suffix": "%",
                    "key": "ttr_lt_15m_perc",
                },
                {
                    "filter": floatformat,
                    "filter_args": 1,
                    "suffix": "%",
                    "key": "ttr_lt_30m_perc",
                },
                {
                    "filter": floatformat,
                    "filter_args": 1,
                    "suffix": "%",
                    "key": "ttr_lt_1h_perc",
                },
                {
                    "filter": floatformat,
                    "filter_args": 1,
                    "suffix": "%",
                    "key": "ttr_lt_3h_perc",
                },
                {
                    "filter": floatformat,
                    "filter_args": 1,
                    "suffix": "%",
                    "key": "ttr_lt_1d_perc",
                },
                {
                    "filter": floatformat,
                    "filter_args": 1,
                    "suffix": "%",
                    "key": "ttr_gte_1d_perc",
                },
            ],
        },
    }

    incident_ttr_by_priority_table = {
        "cols": ["Priority", "< 15m", "< 30m", "< 1h", "< 3h", "< 1d", "≥ 1d"],
        "body": {
            "objects": incident_ttr_by_priority,
            "format": [
                {"filter": None, "suffix": None, "key": "name"},
                {"filter": floatformat, "key": "ttr_lt_15m"},
                {"filter": floatformat, "key": "ttr_lt_30m"},
                {"filter": floatformat, "key": "ttr_lt_1h"},
                {"filter": floatformat, "key": "ttr_lt_3h"},
                {"filter": floatformat, "key": "ttr_lt_1d"},
                {"filter": floatformat, "key": "ttr_gte_1d"},
            ],
        },
        "footer": incident_ttr_by_priority_total_row,
    }

    return {
        "incident_by_priority_table": incident_by_priority_table,
        "incident_age_by_priority_table": incident_age_by_priority_table,
        "incident_ttr_by_priority_percentage_table": incident_ttr_by_priority_percentage_table,
        "incident_ttr_by_priority_table": incident_ttr_by_priority_table,
        "incident_by_priority": incident_by_priority,
        "incident_ttr_by_priority": incident_ttr_by_priority,
        "incident_ttr_by_priority_total_row": incident_ttr_by_priority_total_row,
        "incident_age_by_priority": incident_age_by_priority,
        "incident_by_status_chart": incident_by_status_chart,
        "incident_by_domain_chart": incidents_by_domain_chart,
        "incident_by_priority_chart": incidents_by_priority_chart,
        "page_title": "Reports",
        "date_gte": date_gte,
        "date_lte": date_lte,
        "date_str": date_str,
        "date_input": date_input,
    }


def get_incident_age_by_priority(
    incidents: QuerySet[Incident], date_lte: datetime | None, date_gte: datetime | None
) -> QuerySet[Priority]:
    j_3 = timezone.now() + relativedelta(days=-3)
    j_5 = timezone.now() + relativedelta(days=-5)
    j_7 = timezone.now() + relativedelta(days=-7)
    j_15 = timezone.now() + relativedelta(days=-15)
    j_30 = timezone.now() + relativedelta(days=-30)
    j_90 = timezone.now() + relativedelta(days=-90)

    incident_age_by_priority_q = Q(incident___status__lt=IncidentStatus.CLOSED) & Q(
        incident__in=incidents
    )

    incident_age_by_priority = Priority.objects.order_by("value").annotate(
        incident_0_3=Count(
            "incident",
            filter=(incident_age_by_priority_q & Q(incident__created_at__gt=j_3)),
            output_field=IntegerField(),
        ),
        incident_3_5=Count(
            "incident",
            filter=(
                incident_age_by_priority_q
                & Q(incident__created_at__lt=j_3)
                & Q(incident__created_at__gte=j_5)
            ),
            output_field=IntegerField(),
        ),
        incident_5_7=Count(
            "incident",
            filter=(
                incident_age_by_priority_q
                & Q(incident__created_at__lt=j_5)
                & Q(incident__created_at__gte=j_7)
            ),
            output_field=IntegerField(),
        ),
        incident_7_15=Count(
            "incident",
            filter=(
                incident_age_by_priority_q
                & Q(incident__created_at__lt=j_7)
                & Q(incident__created_at__gte=j_15)
            ),
            output_field=IntegerField(),
        ),
        incident_15_30=Count(
            "incident",
            filter=(
                incident_age_by_priority_q
                & Q(incident__created_at__lt=j_15)
                & Q(incident__created_at__gte=j_30)
            ),
            output_field=IntegerField(),
        ),
        incident_30_90=Count(
            "incident",
            filter=(
                incident_age_by_priority_q
                & Q(incident__created_at__lt=j_30)
                & Q(incident__created_at__gte=j_90)
            ),
            output_field=IntegerField(),
        ),
        incident_gt_90=Count(
            "incident",
            filter=(incident_age_by_priority_q & Q(incident__created_at__lte=j_90)),
            output_field=IntegerField(),
        ),
        incident_0_3_label=Value(""),
        incident_3_5_label=Value(""),
        incident_5_7_label=Value(""),
        incident_7_15_label=Value(""),
        incident_15_30_label=Value(""),
        incident_30_90_label=Value(""),
        incident_gt_90_label=Value(""),
    )

    common_params = f"&_status__lt={IncidentStatus.CLOSED.value}"
    date_gte_param = f"{f'&created__lte={date_lte}' if date_lte else ''}"
    # pyright: reportGeneralTypeIssues=false
    for sev in incident_age_by_priority:
        sev.incident_0_3_label = f'<a class="{"underline" if sev.incident_0_3 else "opacity-50"}" href="/incident/?priority={sev.id}{get_created_at_filter(j_3, None, date_gte)}{common_params}">{sev.incident_0_3}</a>'
        sev.incident_3_5_label = f'<a class="{"underline" if sev.incident_3_5 else "opacity-50"}" href="/incident/?priority={sev.id}{get_created_at_filter(j_5, j_3, date_gte)}{common_params}">{sev.incident_3_5}</a>'
        sev.incident_5_7_label = f'<a class="{"underline" if sev.incident_5_7 else "opacity-50"}" href="/incident/?priority={sev.id}{get_created_at_filter(j_7, j_15, date_gte)}{common_params}">{sev.incident_5_7}</a>'
        sev.incident_7_15_label = f'<a class="{"underline" if sev.incident_7_15 else "opacity-50"}" href="/incident/?priority={sev.id}{get_created_at_filter(j_15, j_7, date_gte)}{common_params}">{sev.incident_7_15}</a>'
        sev.incident_15_30_label = f'<a class="{"underline" if sev.incident_15_30 else "opacity-50"}" href="/incident/?priority={sev.id}{get_created_at_filter(j_30, j_15, date_gte)}{common_params}">{sev.incident_15_30}</a>'
        sev.incident_30_90_label = f'<a class="{"underline" if sev.incident_30_90 else "opacity-50"}" href="/incident/?priority={sev.id}{get_created_at_filter(j_90, j_30, date_gte)}{common_params}">{sev.incident_30_90}</a>'
        sev.incident_gt_90_label = f'<a class="{"underline" if sev.incident_gt_90 else "opacity-50"}" href="/incident/?priority={sev.id}&created_at=2000+-+{str(j_90).replace("-", "/")}{date_gte_param}{common_params}">{sev.incident_gt_90}</a>'

    return incident_age_by_priority


def get_created_at_filter(
    date_gte: datetime, date_lt: datetime | None, date_gte_global: datetime | None
) -> str:
    return f'&created_at={str(get_biggest_date(date_gte_global, date_gte).isoformat()).replace("-", "/").replace("+", "%2B")}+-+{str(date_lt.isoformat()).replace("-", "/").replace("+", "%2B") if date_lt else "now"}'


class _IncidentPriorityAnnotation(TypedDict):
    incident_total: int
    incident_closed: float
    incident_open: float
    incident_closed_percentage: NotRequired[float]


def get_incident_by_priority(
    incidents: QuerySet[Incident],
) -> tuple[
    QuerySet[WithAnnotations[Priority, _IncidentPriorityAnnotation]],
    dict[str, str | float | int],
]:
    incident_by_priority = (
        Priority.objects.order_by("value")
        .filter(incident__in=incidents)
        .annotate(
            incident_total=Count("incident"),  # (len(incidents)),
            incident_closed=Count(
                "incident",
                filter=Q(incident___status=IncidentStatus.CLOSED),
                output_field=FloatField(),
            ),
            incident_open=Count(
                "incident",
                filter=Q(incident___status__lt=IncidentStatus.CLOSED),
                output_field=FloatField(),
            ),
        )
        .filter(incident_total__gt=0.0)
        .annotate(
            incident_closed_percentage=(
                Cast(F("incident_closed"), FloatField())
                / Cast(F("incident_total"), FloatField())
                * 100.0
            )
        )
    )

    # Add the total row
    incident_by_priority_total_row: dict[str, str | float | int] = {"name": "Total"}
    for key in [
        "incident_closed",
        "incident_open",
        "incident_total",
    ]:
        incident_by_priority_total_row[key] = reduce(
            (lambda tot, x: tot + x.__dict__[key]), incident_by_priority, 0
        )

    if incident_by_priority_total_row.get("incident_total") != 0:
        closed = int(incident_by_priority_total_row.get("incident_closed", 0))
        total = incident_by_priority_total_row.get("incident_total")
        if not isinstance(closed, int) or not isinstance(total, int):
            raise ValueError("Invalid type for closed or total")
        incident_by_priority_total_row["incident_closed_percentage"] = (
            closed / total * 100
        )
    return incident_by_priority, incident_by_priority_total_row  # type: ignore[return-value]


class _IncidentByDomainAnnotation(TypedDict):
    incidents_nb: int | float


def get_incidents_by_domain(
    incidents: QuerySet[Incident],
) -> QuerySet[WithAnnotations[Group, _IncidentByDomainAnnotation]]:
    return Group.objects.filter(component__incident__in=incidents).annotate(
        incidents_nb=Count("component__incident", output_field=FloatField()),
    )


def get_incident_time_to_by_priority(
    incidents: QuerySet[Incident], time_to: str = "time_to_fix"
) -> tuple[dict[str, str], QuerySet[Priority]]:
    ttr_q = (
        Q(incident___status__gte=IncidentStatus.FIXED)
        & Q(metric_type__type=time_to)
        & Q(duration__gte=timedelta(seconds=1))
        & Q(incident__in=incidents)
    )
    metric_subquery = (
        IncidentMetric.objects.filter(ttr_q)
        .values("incident__priority")
        .annotate(
            incident_total=Count("incident", output_field=FloatField()),
            ttr_lt_15m=Count(
                "incident",
                filter=Q(duration__lt=timedelta(minutes=15)),
                output_field=FloatField(),
            ),
            ttr_lt_30m=Count(
                "incident",
                filter=Q(duration__lt=timedelta(minutes=30)),
                output_field=FloatField(),
            ),
            ttr_lt_1h=Count(
                "incident",
                filter=Q(duration__lt=timedelta(minutes=60)),
                output_field=FloatField(),
            ),
            ttr_lt_3h=Count(
                "incident",
                filter=Q(duration__lt=timedelta(hours=3)),
                output_field=FloatField(),
            ),
            ttr_lt_1d=Count(
                "incident",
                filter=Q(duration__lt=timedelta(days=1)),
                output_field=FloatField(),
            ),
            ttr_gte_1d=Count(
                "incident",
                filter=Q(duration__gte=timedelta(days=1)),
                output_field=FloatField(),
            ),
        )
    )

    incident_ttr_by_priority = (
        Priority.objects.order_by("value")
        .annotate(
            incident_total=Subquery(
                metric_subquery.filter(incident__priority=OuterRef("pk")).values(
                    "incident_total"
                )[:1]
            ),
            ttr_lt_15m=Subquery(
                metric_subquery.filter(incident__priority=OuterRef("pk")).values(
                    "ttr_lt_15m"
                )[:1]
            ),
            ttr_lt_30m=Subquery(
                metric_subquery.filter(incident__priority=OuterRef("pk")).values(
                    "ttr_lt_30m"
                )[:1]
            ),
            ttr_lt_1h=Subquery(
                metric_subquery.filter(incident__priority=OuterRef("pk")).values(
                    "ttr_lt_1h"
                )[:1]
            ),
            ttr_lt_3h=Subquery(
                metric_subquery.filter(incident__priority=OuterRef("pk")).values(
                    "ttr_lt_3h"
                )[:1]
            ),
            ttr_lt_1d=Subquery(
                metric_subquery.filter(incident__priority=OuterRef("pk")).values(
                    "ttr_lt_1d"
                )[:1]
            ),
            ttr_gte_1d=Subquery(
                metric_subquery.filter(incident__priority=OuterRef("pk")).values(
                    "ttr_gte_1d"
                )[:1]
            ),
        )
        .filter(incident_total__gt=0.0)
        .annotate(
            ttr_lt_15m_perc=(
                Cast(F("ttr_lt_15m"), FloatField())
                / Cast(F("incident_total"), FloatField())
                * 100.0
            ),
            ttr_lt_30m_perc=(
                Cast(F("ttr_lt_30m"), FloatField())
                / Cast(F("incident_total"), FloatField())
                * 100.0
            ),
            ttr_lt_1h_perc=(
                Cast(F("ttr_lt_1h"), FloatField())
                / Cast(F("incident_total"), FloatField())
                * 100.0
            ),
            ttr_lt_3h_perc=(
                Cast(F("ttr_lt_3h"), FloatField())
                / Cast(F("incident_total"), FloatField())
                * 100.0
            ),
            ttr_lt_1d_perc=(
                Cast(F("ttr_lt_1d"), FloatField())
                / Cast(F("incident_total"), FloatField())
                * 100.0
            ),
            ttr_gte_1d_perc=(
                Cast(F("ttr_gte_1d"), FloatField())
                / Cast(F("incident_total"), FloatField())
                * 100.0
            ),
        )
    )
    incident_ttr_by_priority_total_row = {"name": "Total"}
    for key in [
        "ttr_lt_15m",
        "ttr_lt_30m",
        "ttr_lt_1h",
        "ttr_lt_3h",
        "ttr_lt_1d",
        "ttr_gte_1d",
        "incident_total",
    ]:
        incident_ttr_by_priority_total_row[key] = str(
            reduce(
                (lambda total, x: total + x.__dict__[key]), incident_ttr_by_priority, 0
            )
        )
    return incident_ttr_by_priority_total_row, incident_ttr_by_priority


def get_incident_by_status_chart_data(
    incidents: QuerySet[Incident],
) -> dict[str, list[Any]]:
    incident_by_status = incidents.aggregate(
        incident_open=Count(
            "_status",
            filter=Q(_status=IncidentStatus.OPEN),
            output_field=FloatField(),
        ),
        incident_investigating=Count(
            "_status",
            filter=Q(_status=IncidentStatus.INVESTIGATING),
            output_field=FloatField(),
        ),
        incident_fixing=Count(
            "_status",
            filter=Q(_status=IncidentStatus.FIXING),
            output_field=FloatField(),
        ),
        incident_fixed=Count(
            "_status",
            filter=Q(_status=IncidentStatus.FIXED),
            output_field=FloatField(),
        ),
        incident_postmortem=Count(
            "_status",
            filter=Q(_status=IncidentStatus.POST_MORTEM),
            output_field=FloatField(),
        ),
        incident_closed=Count(
            "_status",
            filter=Q(_status=IncidentStatus.CLOSED),
            output_field=FloatField(),
        ),
        incident_total=Count(
            "_status",
            output_field=FloatField(),
        ),
    )
    if incident_by_status.get("incident_total") == 0:
        return {
            "keys": [],
            "values": [],
        }
    val = list(incident_by_status.values())
    val.pop(6)
    return {
        "keys": list(IncidentStatus.labels),
        "values": val,
    }


def get_date_range_from_parameters(
    request: HttpRequest,
) -> tuple[datetime | None, datetime | None, str | None, str | None]:
    unparsed_date = request.GET.get("created_at")
    if not unparsed_date:
        return None, None, None, None

    return get_date_range_from_special_date(unparsed_date)



from __future__ import annotations

from typing import Any
from unittest.mock import MagicMock

import pytest
from slack_sdk.errors import SlackApiError

from firefighter.slack.models.user import SlackUser, SlackUserManager
from firefighter.slack.slack_app import SlackApp
from tests.test_slack.conftest import MockWebClient

# Mock data for testing
mock_slack_response = {
    "user": {
        "id": "U123456",
        "name": "testuser",
        "deleted": False,
        "profile": {
            "real_name": "Test User",
            "first_name": "Test",
            "last_name": "User",
            "email": "testuser@example.com",
            "image_192": "https://example.com/image192.png",
            "image_512": "https://example.com/image512.png",
        },
        "is_bot": False,
    },
    "ok": True,
}


@pytest.fixture
def valid_user_info() -> dict[str, dict[str, Any] | Any]:
    return {
        "ok": True,
        "user": {
            "id": "U01234567",
            "name": "testuser",
            "deleted": False,
            "is_bot": False,
            "profile": {
                "real_name": "Test User",
                "first_name": "Test",
                "last_name": "User",
                "email": "testuser@example.com",
                "image_512": "https://example.com/image_512.jpg",
                "image_192": "https://example.com/image_192.jpg",
            },
        },
    }


def test_unpack_user_info() -> None:
    manager = SlackUserManager()
    unpacked_user_info = manager.unpack_user_info(mock_slack_response)

    assert unpacked_user_info["slack_id"] == "U123456"
    assert unpacked_user_info["username"] == "testuser"
    assert unpacked_user_info["first_name"] == "Test"
    assert unpacked_user_info["last_name"] == "User"
    assert unpacked_user_info["email"] == "testuser@example.com"
    assert unpacked_user_info["image"] == "https://example.com/image512.png"
    assert unpacked_user_info["deleted"] is False


@pytest.mark.django_db
def test_slack_user_link(slack_user_saved: SlackUser):
    link = slack_user_saved.link

    assert link.startswith("slack://user?team=")
    assert f"&id={slack_user_saved.slack_id}" in link


@pytest.mark.django_db
def test_update_user_info(slack_user_saved: SlackUser, mock_web_client: MockWebClient):
    slack_user = slack_user_saved

    # Mock WebClient.users_info
    mock_users_info = MagicMock(return_value=mock_slack_response)
    mock_web_client.users_info = mock_users_info
    slack_user.update_user_info(client=mock_web_client)

    # Assert that users_info was called with the correct slack_id
    mock_users_info.assert_called_once_with(user=slack_user.slack_id)

    # Assert the database values have been updated
    updated_slack_user = SlackUser.objects.get(slack_id=slack_user.slack_id)
    updated_user = updated_slack_user.user

    assert updated_slack_user.username == "testuser"
    assert updated_user.name == "Test User"
    assert updated_user.first_name == "Test"
    assert updated_user.last_name == "User"
    assert updated_user.email == "testuser@example.com"
    assert updated_user.is_active is True
    assert updated_slack_user.image == "https://example.com/image512.png"

    assert updated_slack_user.username != slack_user.username
    assert updated_user.name != slack_user.user.name
    assert updated_user.first_name != slack_user.user.first_name
    assert updated_user.last_name != slack_user.user.last_name
    assert updated_user.email != slack_user.user.email


@pytest.mark.django_db
def test_update_user_info_slack_api_error(
    slack_user_saved: SlackUser, mock_web_client: MockWebClient
):
    slack_user = slack_user_saved
    error_response = MagicMock()
    error_response.status_code = 404
    # Mock WebClient.users_info to raise a SlackApiError
    mock_users_info = MagicMock(
        side_effect=SlackApiError("SlackApiError", response=error_response)
    )
    mock_web_client.users_info = mock_users_info
    slack_user.update_user_info(client=mock_web_client)

    # Assert that users_info was called with the correct slack_id
    mock_users_info.assert_called_once_with(user=slack_user.slack_id)

    # Assert the database values have not been updated
    updated_slack_user = SlackUser.objects.get(slack_id=slack_user.slack_id)
    updated_user = updated_slack_user.user

    assert updated_slack_user.username == slack_user.username
    assert updated_user.name == slack_user.user.name
    assert updated_user.first_name == slack_user.user.first_name
    assert updated_user.last_name == slack_user.user.last_name
    assert updated_user.email == slack_user.user.email
    assert updated_user.is_active is True
    assert updated_slack_user.image == slack_user.image


@pytest.mark.django_db
def test_update_user_info_empty_response(
    slack_user_saved: SlackUser, mock_web_client: MockWebClient
):
    slack_user = slack_user_saved

    # Mock WebClient.users_info to return an empty response
    mock_users_info = MagicMock(return_value={})
    mock_web_client.users_info = mock_users_info
    slack_user.update_user_info(client=mock_web_client)

    # Assert that users_info was called with the correct slack_id
    mock_users_info.assert_called_once_with(user=slack_user.slack_id)

    # Assert the database values have not been updated
    updated_slack_user = SlackUser.objects.get(slack_id=slack_user.slack_id)
    updated_user = updated_slack_user.user

    assert updated_slack_user.username == slack_user.username
    assert updated_user.name == slack_user.user.name
    assert updated_user.first_name == slack_user.user.first_name
    assert updated_user.last_name == slack_user.user.last_name
    assert updated_user.email == slack_user.user.email
    assert updated_user.is_active is True
    assert updated_slack_user.image == slack_user.image


# Add these test functions to your test file


@pytest.mark.django_db
def test_unpack_user_info_bot(valid_user_info: dict[str, Any]):
    valid_user_info["user"]["is_bot"] = True
    unpacked_user_info = SlackUser.objects.unpack_user_info(valid_user_info)
    assert unpacked_user_info["email"] == valid_user_info["user"]["name"]
    assert unpacked_user_info["name"] == valid_user_info["user"]["profile"]["real_name"]


@pytest.mark.django_db
def test_unpack_user_info_no_image(valid_user_info):
    valid_user_info["user"]["profile"]["image_512"] = None
    valid_user_info["user"]["profile"]["image_192"] = None
    unpacked_user_info = SlackUser.objects.unpack_user_info(valid_user_info)
    assert "image" not in unpacked_user_info


@pytest.mark.django_db
def test_link(slack_user_saved: SlackUser):
    slack_user = slack_user_saved
    expected_link = (
        f"slack://user?team={SlackApp().details['team_id']}&id={slack_user.slack_id}"
    )
    assert slack_user.link == expected_link


@pytest.mark.django_db
def test_str_representation(slack_user_saved: SlackUser):
    slack_user = slack_user_saved
    assert str(slack_user) == slack_user.slack_id



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Never

from django.conf import settings
from django.dispatch.dispatcher import receiver

from firefighter.jira_app.client import JiraAPIError, JiraUserNotFoundError
from firefighter.raid.client import client
from firefighter.raid.models import JiraTicket
from firefighter.raid.service import get_jira_user_from_user
from firefighter.slack.messages.slack_messages import (
    SlackMessageIncidentDeclaredAnnouncement,
)
from firefighter.slack.signals import incident_channel_done

if TYPE_CHECKING:
    from firefighter.incidents.models.incident import Incident
    from firefighter.slack.models.incident_channel import IncidentChannel

logger = logging.getLogger(__name__)
RAID_DEFAULT_JIRA_QRAFT_USER_ID: str = settings.RAID_DEFAULT_JIRA_QRAFT_USER_ID
APP_DISPLAY_NAME: str = settings.APP_DISPLAY_NAME


@receiver(signal=incident_channel_done)
def create_ticket(
    sender: Any, incident: Incident, channel: IncidentChannel, **kwargs: Never
) -> JiraTicket:
    # pylint: disable=unused-argument

    jira_user = get_jira_user_from_user(incident.created_by)
    account_id = jira_user.id
    # XXX Better description
    # XXX Custom field with FireFighter ID/link?
    # XXX Set affected environment custom field
    # XXX Set custom field impacted area to group/domain?
    priority: int = incident.priority.value if 1 <= incident.priority.value <= 4 else 1
    issue = client.create_issue(
        issuetype="Incident",
        summary=incident.title,
        description=f"""{incident.description}\n
\n
🧯 This incident has been created for a critical incident. Links below to Slack and {APP_DISPLAY_NAME}.\n
📦 Component: {incident.component.name} ({incident.component.group.name})\n
{incident.priority.emoji} Priority: {incident.priority.name}\n""",
        assignee=None,
        reporter=account_id,
        priority=priority,
    )
    issue_id = issue.get("id")
    if issue_id is None:
        logger.error(f"Could not create Jira ticket for incident {incident.id}")
        raise JiraAPIError("Could not create Jira ticket")
    try:
        default_jira_user = client.get_jira_user_from_jira_id(
            RAID_DEFAULT_JIRA_QRAFT_USER_ID
        )
    except JiraUserNotFoundError:
        logger.exception(
            f"Could not find Jira default reporter user with account id {RAID_DEFAULT_JIRA_QRAFT_USER_ID}"
        )
    # Add watcher reporter
    try:
        client.jira.add_watcher(issue=issue_id, watcher=account_id)
    except JiraAPIError:
        logger.exception(
            f"Could not add the watcher with account id {account_id} to the ticket {issue_id}"
        )
        # Removing default watcher TeamQraft
        try:
            client.jira.remove_watcher(issue=issue_id, watcher=default_jira_user)
        except JiraAPIError:
            logger.exception(
                f"Could not remove the watcher {default_jira_user} to the ticket {issue_id}"
            )

    client.jira.add_simple_link(
        issue=str(issue_id),
        object={
            "url": incident.status_page_url,
            "title": f"{APP_DISPLAY_NAME} incident #{incident.id}",
        },
    )

    client.jira.add_simple_link(
        issue=str(issue_id),
        object={"url": channel.link, "title": f"Slack conversation #{channel.name}"},
    )
    issue["business_impact"] = issue.get("business_impact", "")
    jira_ticket = JiraTicket.objects.create(**issue, incident=incident)

    channel.add_bookmark(
        title="Jira ticket",
        link=jira_ticket.url,
        emoji=":jira_new:",
    )

    channel.send_message_and_save(SlackMessageIncidentDeclaredAnnouncement(incident))

    return jira_ticket



from __future__ import annotations

import os
from typing import Any

import gunicorn.app.wsgiapp
from gunicorn.app.base import Application

HTTP_BIND = os.getenv("HTTP_BIND", "0.0.0.0")  # noqa: S104
HTTP_PORT = int(os.getenv("HTTP_PORT", "8000"))


class StandaloneApplication(gunicorn.app.wsgiapp.WSGIApplication):
    def __init__(self, app: str, options: dict[str, Any]) -> None:  # noqa: ARG002
        self.options = options
        self.app_uri = "firefighter.firefighter.wsgi"
        super().__init__()

    def load_config(self) -> None:
        Application.load_config(self)
        if self.cfg is None:
            raise ValueError("cfg is None")
        self.cfg.set("workers", int(os.getenv("GUNICORN_WORKERS", "8")))
        self.cfg.set("bind", f"{HTTP_BIND}:{HTTP_PORT}")

        if "GUNICORN_TIMEOUT" in os.environ:
            self.cfg.set("timeout", int(os.environ["GUNICORN_TIMEOUT"]))
        self.app_uri = "firefighter.firefighter.wsgi"


def main() -> None:
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "firefighter.firefighter.settings")
    # ruff: noqa: PLC0415
    from django import setup

    setup()
    from django.core.management import call_command

    call_command("migrate")
    call_command("collectstatic", "--no-input")

    StandaloneApplication(
        "%(prog)s [OPTIONS] [APP_MODULE]",
        {
            "bind": f"{HTTP_BIND}:{HTTP_PORT}",  # noqa: S104
            "workers": 4,
            "wsgi_app": "firefighter.firefighter.wsgi",
        },
    ).run()


if __name__ == "__main__":
    main()



# Generated by Django 4.2.7 on 2023-12-04 11:09


import logging

from django.db import migrations, models

logger = logging.getLogger(__name__)

# Mapping for updating ImpactLevel names based on value and impact_type
IMPACT_LEVEL_NAME_MAPPING = {
    # Format: (value, impact_type): (new_name, description)
    ("LT", "business_impact"): (
        "Business Lowest Not measurable",
        "Lowest business impact, not measurable."
    ),
    ("LO", "business_impact"): (
        "Business Lower Not measurable",
        "Lower business impact, not measurable."
    ),
    ("MD", "business_impact"): (
        "Uncertain Business Impact",
        "Uncertain business value daily loss."
    ),
    ("HI", "business_impact"): (
        "Major Business Impact (<5% BV daily loss)",
        "Major business impact, less than 5% business value daily loss."
    ),
    ("HT", "business_impact"): (
        "Critical Business Impact (>5% BV daily loss)",
        "Critical business impact, more than 5% business value daily loss."
    ),
    ("NO", "business_impact"): (
        "No impact on business.",
        ""
    ),
    ("LT", "customers_impact"): (
        "Cosmetic issues for few customers",
        "Minor functional affecting few users / cosmetic issues."
    ),
    ("LO", "customers_impact"): (
        "Minor functional issues for many customers",
        "Minor functional issues affecting many customers."
    ),
    ("MD", "customers_impact"): (
        "Major functional issues for many customers",
        "Many customer with major functional issues (no affecting BV)."
    ),
    ("HI", "customers_impact"): (
        "Some customers have major issues",
        "Purchase flow experience with major experience issues impacting <5% of customers.\nAccess to product not working affecting <5% of customers/errors\nMarketplace unstable <30% errors."
    ),
    ("HT", "customers_impact"): (
        "Critical issue for many customers",
        "Purchase flow experience with major experience issues impacting >5% of customers.\nAccess to product not working affecting >5% of customers/errors\nMarketplace unstable >30% errors."
    ),
    ("NO", "customers_impact"): (
        "No impact for customers.",
        ""
    ),
    ("LT", "sellers_impact"): (
        "Cosmetic issues for few sellers",
        "Minor cosmetic issues affecting few sellers."
    ),
    ("LO", "sellers_impact"): (
        "Minor functional issues for many sellers",
        "Minor functional issues affecting many sellers."
    ),
    ("MD", "sellers_impact"): (
        "Major functional issues for many sellers",
        "Many sellers with major functional issues (no affecting fulfillment / catalog availability / order).\nIssues affecting commercial operations (e.g. wrong pricing on inventory / offers / etc) of big part of our inventory (e.g. big / important sellers)."
    ),
    ("HI", "sellers_impact"): (
        "Some sellers have major issues",
        "Toolbox access hindered affecting <20% of sellers.\nImpossibility of publishing / modifying catalogue affecting <20% of sellers."
    ),
    ("HT", "sellers_impact"): (
        "Critical issue for many sellers",
        "Toolbox access hindered affecting >20% of sellers.\nImpossibility of publishing / modifying catalogue affecting >20% of sellers."
    ),
    ("NO", "sellers_impact"): (
        "No impact for sellers.",
        ""
    ),
    ("LT", "employees_impact"): (
        "Degraded service for MM workers or partners but not blocked",
        "Service is degraded for MM workers or partners, but they are not blocked."
    ),
    ("LO", "employees_impact"): (
        "Few MM workers or partners affected systematically or blocked",
        "A few MM workers or partners are systematically affected or blocked."
    ),
    ("MD", "employees_impact"): (
        "Full departments blocked on their work.",
        "Entire departments are blocked in their work."
    ),
    ("NO", "employees_impact"): (
        "No impact for employee.",
        ""
    ),
}


def update_impact_level_names(apps, _schema_editor):
    ImpactLevel = apps.get_model("incidents", "ImpactLevel")
    ImpactType = apps.get_model("incidents", "ImpactType")
    for (value, impact_type_value), (new_name, description) in IMPACT_LEVEL_NAME_MAPPING.items():
        try:
            impact_type = ImpactType.objects.filter(value=impact_type_value).first()
            if not impact_type:
                logger.warning(f"ImpactType '{impact_type}' does not exist. Skipping update for value '{value}' and impact_type '{impact_type_value}'.")
                continue

            impact_level = ImpactLevel.objects.get(value=value, impact_type=impact_type.id)
            impact_level.name = new_name
            impact_level.description = description
            impact_level.save()
            logger.info(f"Updated ImpactLevel ({value}, {impact_type}) to name '{new_name}'")
        except ImpactLevel.DoesNotExist:
            logger.warning(f"ImpactLevel with value '{value}' and impact_type '{impact_type}' does not exist.")


class Migration(migrations.Migration):

    dependencies = [
        ("incidents", "0014_update_components_slack_groups"),
    ]

    operations = [
        migrations.AddField(
            model_name="impactlevel",
            name="description",
            field=models.TextField(null=True, blank=True),
        ),
        migrations.RunPython(update_impact_level_names),
    ]



from __future__ import annotations

import datetime
import logging
import re
from typing import TYPE_CHECKING, Any, cast

from django.core.cache import cache as dj_cache
from django.utils import timezone
from slack_sdk.models.blocks import Block, ContextBlock, HeaderBlock, SectionBlock
from slack_sdk.models.blocks.basic_components import MarkdownTextObject
from slack_sdk.models.blocks.block_elements import ButtonElement

from firefighter.incidents.forms.update_key_events import IncidentUpdateKeyEventsForm
from firefighter.slack.messages.base import SlackMessageStrategy, SlackMessageSurface
from firefighter.slack.views.modals.base_modal.base import MessageForm

if TYPE_CHECKING:
    from django_redis.client.default import DefaultClient
    from slack_bolt.context.ack.ack import Ack

    from firefighter.incidents.models.incident import Incident, User
    from firefighter.slack.views.modals.base_modal.form_utils import SlackForm

cache = cast("DefaultClient", dj_cache)
logger = logging.getLogger(__name__)

# XXX Remove legacy `post_mortem_milestone` once we're sure it's not used anymore
MILESTONE_ID_REGEX = re.compile(
    r"^(post_mortem_milestone|key_event)_?(?P<type>date|time|clear)?_(?P<id>.*)$"
)

TZ = timezone.get_current_timezone()


class KeyEvents(MessageForm[IncidentUpdateKeyEventsForm]):
    form_class = IncidentUpdateKeyEventsForm
    callback_id = MILESTONE_ID_REGEX
    callback_action = True
    wrapper = "action"

    def build_modal_fn(self, incident: Incident) -> list[Block]:
        slack_form: SlackForm[IncidentUpdateKeyEventsForm] = self.get_form_class()(
            incident=incident
        )

        return self.get_blocks_from_form(slack_form.form)

    def get_blocks_from_form(self, form: IncidentUpdateKeyEventsForm) -> list[Block]:
        blocks: list[Block] = [HeaderBlock(text=":stopwatch:  Key events time")]
        slack_form: SlackForm[IncidentUpdateKeyEventsForm] = self.get_form_class()
        slack_form.form = form
        blocks += slack_form.slack_blocks()
        # XXX Check form.is_valid()
        missing_milestones = form.incident.missing_milestones()
        if len(missing_milestones) > 0:
            blocks.append(
                SectionBlock(
                    text=f":warning: Some required key events are missing: {', '.join(missing_milestones)}. Once all required key events have been submitted, you'll be able to to close this incident."
                )
            )
        else:
            blocks.append(
                # XXX Check if postmortem and change message
                SectionBlock(
                    text=":white_check_mark: All required key events have been submitted. Once you're ready, you can close the incident.",
                    accessory=ButtonElement(
                        text="Close incident",
                        value=str(form.incident.id),
                        action_id="close_incident",
                    ),
                )
            )
        blocks.append(
            ContextBlock(
                elements=[
                    MarkdownTextObject(
                        text=f"Last update: {datetime.datetime.now(tz=timezone.get_current_timezone()).strftime('%Y-%m-%d %H:%M:%S')}"
                    )
                ]
            ),
        )
        return blocks

    def handle_modal_fn(  # type: ignore[override]
        self, ack: Ack, body: dict[str, Any], user: User, incident: Incident
    ) -> None:
        """Handle the time and date inputs for the key events."""
        logger.debug(body)

        slack_form: SlackForm[IncidentUpdateKeyEventsForm] | None = (
            self.handle_form_errors(
                ack,
                body,
                forms_kwargs={
                    "incident": incident,
                    "user": user,
                },
            )
        )
        form = slack_form.form if slack_form else None

        if form is None:
            logger.warning("Form is None, skipping save")
            return
        if len(form.errors) > 0:
            self.update_with_form()
            return
        self.form = form
        self.form.save()
        incident.compute_metrics()

        self.update_with_form()

    def update_with_form(
        self,
    ) -> None:
        self.form.incident.conversation.send_message_and_save(
            SlackMessageKeyEvents(self.form.incident)
        )


key_events_surface = KeyEvents()


class SlackMessageKeyEvents(SlackMessageSurface):
    id = "ff_incident_key_events_form"
    strategy: SlackMessageStrategy = SlackMessageStrategy.UPDATE
    incident: Incident | None
    form: IncidentUpdateKeyEventsForm | None

    def __init__(
        self,
        incident: Incident | None = None,
        form: IncidentUpdateKeyEventsForm | None = None,
    ) -> None:
        if incident is None and form is None:
            raise ValueError("Either incident or form must be set")
        self.incident = incident
        self.form = form
        super().__init__()

    def get_blocks(self) -> list[Block]:
        if self.incident is not None:
            return KeyEvents().build_modal_fn(self.incident)
        if self.form is not None:
            return KeyEvents().get_blocks_from_form(self.form)

        raise ValueError("Either incident or form must be set")

    def get_text(self) -> str:
        if self.incident:
            incident = self.incident
        elif self.form:
            incident = self.form.incident
        else:
            raise ValueError("Either incident or form must be set")
        return f"Please fill the Key events for {incident}."



# Generated by ChatGPT on 2024-06-24
from django.db import migrations


def reorder_impact_types(apps, schema_editor):
    ImpactType = apps.get_model("incidents", "ImpactType")
    # On suppose que les valeurs sont :
    # business_impact, customers_impact, sellers_impact, employees_impact
    order_map = {
        "business_impact": 1,
        "customers_impact": 2,
        "sellers_impact": 3,
        "employees_impact": 4,
    }
    for value, order in order_map.items():
        impact_type = ImpactType.objects.filter(value=value).first()
        if impact_type:
            impact_type.order = order
            impact_type.save(update_fields=["order"])


class Migration(migrations.Migration):
    dependencies = [
        ("incidents", "0016_update_business_incidents_and_level"),
    ]
    operations = [
        migrations.RunPython(reorder_impact_types),
    ]



# Generated by Django 4.2.7 on 2023-12-04 11:09


import logging

from django.db import migrations

logger = logging.getLogger(__name__)


def update_incidents(apps, _schema_editor):
    ImpactLevel = apps.get_model("incidents", "ImpactLevel")
    ImpactType = apps.get_model("incidents", "ImpactType")
    Incident = apps.get_model("incidents", "Incident")

    business_impact_type = ImpactType.objects.filter(value="business_impact").first()
    if not business_impact_type:
        logger.error("ImpactType 'business_impact' does not exist. Skipping incident update.")
        return

    md_impact_level = ImpactLevel.objects.filter(value="MD", impact_type=business_impact_type.id).first()
    if not md_impact_level:
        logger.error("ImpactLevel 'MD' for business_impact does not exist. Skipping incident update.")
        return

    for level_value in ["LT", "LO"]:
        old_impact_level = ImpactLevel.objects.filter(value=level_value, impact_type=business_impact_type.id).first()
        if not old_impact_level:
            logger.error(f"ImpactLevel '{level_value}' for business_impact does not exist. Skipping.")
            continue

        # Find incidents that have the old impact level through the impacts relationship
        incidents_with_old_level = Incident.objects.filter(impacts=old_impact_level.id)
        updated_count = 0

        for incident in incidents_with_old_level:
            # Remove the old impact level and add the new one if not already present
            incident.impacts.remove(old_impact_level)
            if not incident.impacts.filter(id=md_impact_level.id).exists():
                incident.impacts.add(md_impact_level)
            updated_count += 1

        logger.info(f"Updated {updated_count} incidents from impact_level '{level_value}' to 'MD' for business_impact.")


def remove_old_impact_levels(apps, _schema_editor):
    ImpactLevel = apps.get_model("incidents", "ImpactLevel")
    ImpactType = apps.get_model("incidents", "ImpactType")
    Impact = apps.get_model("incidents", "Impact")

    business_impact_type = ImpactType.objects.filter(value="business_impact").first()
    if not business_impact_type:
        logger.error("ImpactType 'business_impact' does not exist. Skipping impact level removal.")
        return

    md_impact_level = ImpactLevel.objects.filter(value="MD", impact_type=business_impact_type.id).first()
    if not md_impact_level:
        logger.error("ImpactLevel 'MD' for business_impact does not exist. Cannot update Impact objects.")
        return

    for level_value in ["LT", "LO"]:
        impact_level = ImpactLevel.objects.filter(value=level_value, impact_type=business_impact_type.id).first()
        if impact_level:
            # Update all Impact objects that reference this impact level
            impacts_to_update = Impact.objects.filter(impact_level=impact_level)
            updated_count = impacts_to_update.update(impact_level=md_impact_level)
            logger.info(f"Updated {updated_count} Impact objects from '{level_value}' to 'MD' for business_impact.")

            # Now safe to delete the impact level
            impact_level.delete()
            logger.info(f"Removed ImpactLevel '{level_value}' for business_impact.")
        else:
            logger.error(f"ImpactLevel '{level_value}' for business_impact not found.")


class Migration(migrations.Migration):

    dependencies = [
        ("incidents", "0015_update_impact_level"),
    ]
    operations = [
        migrations.RunPython(update_incidents),
        migrations.RunPython(remove_old_impact_levels),
    ]



from __future__ import annotations

import uuid
from datetime import timedelta

from django.db import models
from django.db.models import Manager
from django_stubs_ext.db.models import TypedModelMeta


class Priority(models.Model):
    """A priority for an incident."""

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=128, unique=True)
    value = models.IntegerField(unique=True)
    emoji = models.CharField(max_length=5, default="🔥")
    description = models.CharField(max_length=128)
    order = models.IntegerField(default=0, unique=True)
    default = models.BooleanField(default=False)
    needs_postmortem = models.BooleanField(default=False)
    enabled_create = models.BooleanField(
        default=True, help_text="Can you create a new incident with this priority?"
    )
    enabled_update = models.BooleanField(
        default=True, help_text="Can you update an incident to this priority?"
    )
    reminder_time = models.DurationField(
        default=timedelta(hours=1),
        help_text="Time before an incident without any incident update will receive a reminder.",
    )
    sla = models.DurationField(
        verbose_name="SLA",
        default=timedelta(hours=1),
        help_text="Service Level Agreement",
    )
    recommended_response_type = models.CharField(max_length=128, null=True, blank=True)
    # XXX Add choices to recommended_response_type

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    objects = Manager["Priority"]()

    class Meta(TypedModelMeta):
        ordering = ["order"]
        verbose_name_plural = "priorities"
        constraints = [
            models.UniqueConstraint(
                fields=["default"],
                name="unique_priority_default",
                condition=models.Q(default=True),
            ),
        ]

    def __str__(self) -> str:
        return self.name

    @classmethod
    def get_default(cls) -> Priority:
        return Priority.objects.get(default=True)

    @property
    def full_name(self) -> str:
        return f"{self.emoji} {self.name} - {self.description}"



from __future__ import annotations

from firefighter.firefighter.settings.components.common import INSTALLED_APPS
from firefighter.firefighter.settings.settings_utils import config

ENABLE_PAGERDUTY: bool = config("ENABLE_PAGERDUTY", cast=bool, default=False)
"""Enable PagerDuty integration."""

if ENABLE_PAGERDUTY:
    INSTALLED_APPS.append("firefighter.pagerduty")

    PAGERDUTY_API_KEY: str = config("PAGERDUTY_API_KEY")
    """PagerDuty API key."""
    PAGERDUTY_ACCOUNT_EMAIL: str = config("PAGERDUTY_ACCOUNT_EMAIL")
    """PagerDuty account email, linked to the API key."""
    PAGERDUTY_URL: str = config("PAGERDUTY_URL", default="https://api.pagerduty.com")
    """PagerDuty API URL."""



from __future__ import annotations

import logging
from typing import Any, NotRequired, Required, TypedDict, Unpack

from django.utils.safestring import SafeString
from django_components import component
from django_components.slots import SlotFunc

logger = logging.getLogger(__name__)

SlotContent = str | SafeString | SlotFunc[Any]


class Kwargs(TypedDict, total=False):
    autoplay: Required[bool]
    header: NotRequired[bool]


class Data(TypedDict):
    autoplay: bool


class Slots(TypedDict):
    modal_header: NotRequired[SlotContent]
    modal_content: NotRequired[SlotContent]
    modal_enabler: NotRequired[SlotContent]


@component.register("modal")
class Modal(component.Component):
    template_name = "modal/modal.html"

    def get_context_data(self, *args: Any, **kwargs: Unpack[Kwargs]) -> Data:
        return Kwargs(autoplay=kwargs["autoplay"])



from __future__ import annotations

import logging
from typing import Any, NotRequired, Required, TypedDict

from django_components import component

from firefighter.incidents.models.user import User

logger = logging.getLogger(__name__)


class Data(TypedDict):
    user: User
    size_tailwind: int
    size_px: int


Args = tuple[User]


class Kwargs(TypedDict, total=False):
    user: Required[User]
    size: NotRequired[str]


@component.register("avatar")
class Avatar(component.Component):
    template_name = "avatar/avatar.html"

    def get_context_data(self, user: User, **kwargs: Any) -> Data:
        size_px, size_tailwind = (40, 10) if kwargs.get("size") == "md" else (80, 20)
        return {
            "user": user,
            "size_tailwind": size_tailwind,
            "size_px": size_px,
        }



from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING, Any, Generic, TypeVar, cast

from django import forms
from django.db.models import Model
from slack_sdk.models.views import View

from firefighter.firefighter.utils import get_in
from firefighter.incidents.forms.create_incident import CreateIncidentFormBase
from firefighter.incidents.models.priority import Priority
from firefighter.incidents.signals import create_incident_conversation
from firefighter.slack.views.modals.base_modal.base import ModalForm
from firefighter.slack.views.modals.base_modal.form_utils import (
    SlackFormJSONEncoder,
    slack_view_submission_to_dict,
)
from firefighter.slack.views.modals.base_modal.modal_utils import (
    push_modal,
    update_modal,
)
from firefighter.slack.views.modals.opening.types import OpeningData

if TYPE_CHECKING:
    from slack_bolt import BoltRequest
    from slack_bolt.context.ack.ack import Ack

    from firefighter.incidents.models.incident import Incident
    from firefighter.incidents.models.user import User

logger = logging.getLogger(__name__)


T = TypeVar("T", bound=CreateIncidentFormBase)


class SetIncidentDetails(ModalForm[T], Generic[T]):
    callback_id: str
    push_action: str
    title: str = "Open an incident"
    submit_text: str = "Save details"
    form_class: type[T]
    id: str

    def __init__(self) -> None:
        if hasattr(self, "id"):
            self.push_action = f"push_{self.id}"
        super().__init__()

    def build_modal_fn(
        self, open_incident_context: OpeningData | None = None, **kwargs: Any
    ) -> View:
        open_incident_context = open_incident_context or {}
        details_form_data: dict[str, Any] = (
            open_incident_context.get("details_form_data") or {}
        )
        for field_name, field in self.form_class.base_fields.items():
            if (
                isinstance(field, forms.ModelChoiceField)
                and field_name in details_form_data
                and not isinstance(details_form_data[field_name], Model)
                and field.queryset is not None
                and field.queryset.model
                and hasattr(field.queryset.model, "DoesNotExist")
            ):
                try:
                    details_form_data[field_name] = field.queryset.get(
                        pk=details_form_data[field_name]
                    )
                except field.queryset.model.DoesNotExist:
                    details_form_data[field_name] = None

        return View(
            type="modal",
            title=self.title[:24],
            submit=self.submit_text[:24],
            close="Close details",
            callback_id=self.callback_id,
            blocks=self.get_form_class()(initial=details_form_data).slack_blocks(),
            private_metadata=json.dumps(
                open_incident_context, cls=SlackFormJSONEncoder
            ),
        )

    def handle_modal_fn(  # type: ignore
        self,
        ack: Ack,
        body: dict[str, Any],
        user: User | None = None,
    ):
        """Handle response from /incident open modal."""
        private_metadata: dict[str, Any] = json.loads(
            body.get("view", {}).get("private_metadata", {})
        )
        priority = private_metadata.get("details_form_data", {}).get("priority", None)
        if priority is not None:
            priority = Priority.objects.get(pk=priority)

        slack_form = self.get_form_class()(
            data={**slack_view_submission_to_dict(body), "priority": priority}
        )
        form: T = slack_form.form
        if form.is_valid():
            ack()
            skip_form = False
        else:
            ack(
                response_action="errors",
                errors={
                    k: self._concat_validation_errors_msg(v)
                    for k, v in form.errors.as_data().items()
                },
            )
            skip_form = True

        if skip_form:
            return
        # ruff: noqa: PLC0415
        from firefighter.slack.views.modals.open import modal_open

        if "priority" in private_metadata and isinstance(
            private_metadata["priority"], str
        ):
            private_metadata["priority"] = Priority.objects.get(
                pk=private_metadata["priority"]
            )
        data = OpeningData(
            details_form_data=cast("dict[str, Any]", form.data),
            impact_form_data=private_metadata.get("impact_form_data"),
            incident_type=private_metadata.get("incident_type"),
            response_type=private_metadata.get("response_type"),
            priority=private_metadata.get("priority"),
        )
        view = modal_open.build_modal_fn(open_incident_context=data, user=user)
        view.private_metadata = json.dumps(data, cls=SlackFormJSONEncoder)

        update_modal(
            view=view,
            trigger_id=get_in(body, ["trigger_id"]),
            view_id=get_in(body, ["view", "root_view_id"]),
        )

    @staticmethod
    def _trigger_incident_workflow(
        incident: Incident,
    ) -> None:
        create_incident_conversation.send(
            "incident_created",
            incident=incident,
        )

    def _handle_action_push(
        self,
        request: BoltRequest,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        body = request.body
        ack: Ack = request.context.ack
        data: OpeningData = (
            json.loads(body.get("actions", [{}])[0].get("value", {})) or {}  # type: ignore
        )

        view = self.build_modal_fn(open_incident_context=data)
        view.private_metadata = json.dumps(data, cls=SlackFormJSONEncoder)
        ack()
        push_modal(view=view, body=body)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, cast

from django.utils import timezone
from django_filters.views import FilterView
from django_tables2.views import SingleTableMixin

from firefighter.incidents.models import Component
from firefighter.incidents.models.component import ComponentFilterSet
from firefighter.incidents.tables import ComponentsTable

if TYPE_CHECKING:
    from django_tables2.tables import Table

    from firefighter.firefighter.utils import HtmxHttpRequest

logger = logging.getLogger(__name__)

TZ = timezone.get_current_timezone()


class ComponentsViewList(SingleTableMixin, FilterView):
    table_class = ComponentsTable
    context_object_name = "components"
    filterset_class = ComponentFilterSet
    model = Component

    paginate_by = 150
    paginate_orphans = 20

    # XXX Export MTBF in API! Change defaults fields for component
    # XXX Help/documentation on the MTBF calculation
    # XXX MTBF when incidents are far appart ? Check and add a warning

    def get_table(self, **kwargs: Any) -> Table:
        table = super().get_table(**kwargs)
        # Hide MTBF and incident count if no date range is selected (they will be empty)
        if not self.request.GET or not self.request.GET.get("metrics_period"):
            table.exclude = ("mtbf", "incident_count")
        return table

    def get_template_names(self) -> list[str]:
        request = cast("HtmxHttpRequest", self.request)
        if request.htmx and not request.htmx.boosted:
            template_name = "layouts/partials/partial_table_list_paginated.html"
        else:
            template_name = "pages/component_list.html"

        return [template_name]

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """No *args to pass."""
        # Call the base implementation first to get a context
        context = super().get_context_data(**kwargs)
        context["filter_order"] = [
            "search",
            "metrics_period",
            "group",
        ]
        context["page_title"] = "Issue categories list"
        return context



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Final, Never

from django.conf import settings
from drf_spectacular.utils import OpenApiExample, extend_schema
from rest_framework import generics, mixins, permissions, status
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response

from firefighter.raid.models import JiraTicket
from firefighter.raid.serializers import (
    JiraWebhookCommentSerializer,
    JiraWebhookUpdateSerializer,
    LandbotIssueRequestSerializer,
)

RAID_DEFAULT_JIRA_QRAFT_USER_ID: Final[str] = settings.RAID_DEFAULT_JIRA_QRAFT_USER_ID


logger = logging.getLogger(__name__)
if TYPE_CHECKING:
    from rest_framework.request import Request


@extend_schema(
    examples=[
        OpenApiExample(
            "Create an issue",
            summary="Create an issue",
            description="Example of a working request coming from landbot. be careful, as you probably don't have a john.doe user and thus may fallback to the default user.",
            value={
                "summary": "Swagger test",
                "description": "Description test where you want to depict your issue",
                "seller_contract_id": "12345678",
                "zoho": "https://crmplus.zoho.eu/mycrmlink/index.do/cxapp/agent/mycompany/all/tickets/details/123456789",
                "platform": "FR",
                "reporter_email": "john.doe@mycompany.com",
                "impacted_area": "Sellers>MF Stock Replenishment",
                "project": "SBI",
                "labels": ["originBot", "ProductsMerge"],
                "environments": ["PRD", "STG"],
                "issue_type": "Incident",
                "business_impact": "High",
                "priority": 4,
                "attachments": [
                    "https://storage.googleapis.com/media.landbot.io/123456/customers/123456789/ABCDEFGHIJKLMNOPQRSTUVWXYZ123456.png",
                    "https://storage.googleapis.com/media.landbot.io/123456/customers/123456780/ABCDEFGHIJKLMNOPQRSTUVWXYZ123450.png",
                ],
            },
            request_only=True,  # signal that example only applies to requests
            response_only=False,  # signal that example only applies to responses
        ),
        OpenApiExample(
            "Create incident response",
            status_codes=["201"],
            value={"https://mycompany.atlassian.net/browse/1234567"},
            request_only=False,
            response_only=True,
        ),
    ]
)
class CreateJiraBotView(
    mixins.CreateModelMixin,
    generics.GenericAPIView[JiraTicket],
):
    queryset = JiraTicket.objects.all().select_related(
        "id",
        "key",
        "assignee",
        "reporter",
        "watchers",
        "business_impact",
        "description",
        "summary",
        "issue_type",
    )
    serializer_class = LandbotIssueRequestSerializer
    permission_classes = [permissions.AllowAny]
    renderer_classes = [JSONRenderer]

    def post(self, request: Request, *args: Never, **kwargs: Never) -> Response:
        """Allow to create a Jira ticket through Landbot.
        Requires a valid Bearer token, that you can create in the back-office if you have the right permissions.
        """
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        headers = self.get_success_headers(serializer.data)
        return Response(
            serializer.data.get("key"), status=status.HTTP_201_CREATED, headers=headers
        )


class JiraUpdateAlertView(
    generics.CreateAPIView[Any],
):
    serializer_class = JiraWebhookUpdateSerializer
    # XXX: Work on webhook token for authentication_classes
    authentication_classes = []
    permission_classes = [permissions.AllowAny]
    renderer_classes = [JSONRenderer]

    def post(self, request: Request, *args: Never, **kwargs: Never) -> Response:
        """Allow to send a message in Slack when some fields ("Priority", "project", "description", "status") of a Jira ticket are updated.
        Requires a valid Bearer token, that you can create in the back-office if you have the right permissions.
        """
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        return Response()


class JiraCommentAlertView(
    generics.CreateAPIView[Any],
):
    serializer_class = JiraWebhookCommentSerializer
    # XXX: Work on webhook token for authentication_classes
    authentication_classes = []
    permission_classes = [permissions.AllowAny]
    renderer_classes = [JSONRenderer]

    def post(self, request: Request, *args: Never, **kwargs: Never) -> Response:
        """Allow to send a message in Slack when a comment in a Jira ticket is created or modified.
        Requires a valid Bearer token, that you can create in the back-office if you have the right permissions.
        """
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        return Response()



from __future__ import annotations

import logging
import re
from copy import deepcopy
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, cast

import django_filters
from django.apps import apps
from django.conf import settings
from django.contrib.postgres.aggregates import ArrayAgg
from django.contrib.postgres.search import SearchQuery, SearchRank, SearchVector
from django.db import models, transaction
from django.urls import reverse
from django.utils.text import Truncator
from django.utils.timezone import localtime
from django_filters.filters import (
    ModelMultipleChoiceFilter,
    MultipleChoiceFilter,
    OrderingFilter,
)
from django_stubs_ext.db.models import TypedModelMeta
from taggit.managers import TaggableManager

from firefighter.firefighter.fields_forms_widgets import (
    CustomCheckboxSelectMultiple,
    FFDateRangeSingleFilter,
    GroupedCheckboxSelectMultiple,
)
from firefighter.incidents import signals
from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.models.component import Component
from firefighter.incidents.models.environment import Environment
from firefighter.incidents.models.group import Group
from firefighter.incidents.models.incident_membership import (
    IncidentMembership,
    IncidentRole,
)
from firefighter.incidents.models.incident_role_type import IncidentRoleType
from firefighter.incidents.models.incident_update import IncidentUpdate
from firefighter.incidents.models.metric_type import IncidentMetric, MetricType
from firefighter.incidents.models.milestone_type import MilestoneType
from firefighter.incidents.models.priority import Priority
from firefighter.incidents.models.severity import Severity
from firefighter.incidents.models.user import User
from firefighter.incidents.signals import (
    incident_closed,
    incident_created,
    incident_updated,
)

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence  # noqa: F401
    from decimal import Decimal
    from uuid import UUID

    from django.db.models.query import QuerySet
    from django_stubs_ext.db.models.manager import RelatedManager

    from firefighter.incidents.models.impact import (
        Impact,  # noqa: F401
        IncidentImpact,  # noqa: F401
    )
    from firefighter.incidents.models.incident_cost import IncidentCost
    from firefighter.slack.models import IncidentChannel

    if settings.ENABLE_CONFLUENCE:
        from firefighter.confluence.models import PostMortem


NON_ALPHANUMERIC_CHARACTERS = re.compile(r"[^\da-zA-Z]+")

TD0 = timedelta(0)


class IncidentManager(models.Manager["Incident"]):
    def get_or_none(self, **kwargs: Any) -> Incident | None:
        try:
            return self.get(**kwargs)
        except Incident.DoesNotExist:
            return None

    def declare(self, **kwargs: Any) -> Incident:
        """Create an Incident and its first IncidentUpdate.
        Send the incident_created signal.
        Returns the saved incident, no need to .save().
        """
        with transaction.atomic():
            if "private" not in kwargs:
                kwargs["private"] = kwargs["component"].private
            if "severity" not in kwargs and "priority" in kwargs:
                kwargs["severity"] = Severity.objects.get(
                    value=kwargs["priority"].value
                )
            incident: Incident = super().create(**kwargs)
            for required_default_role in IncidentRoleType.objects.filter(required=True):
                incident.roles_set.add(
                    IncidentRole.objects.create(
                        incident=incident,
                        role_type=required_default_role,
                        user=incident.created_by,
                    )
                )

            first_incident_update = IncidentUpdate(
                title=incident.title,
                description=incident.description,
                status=incident.status,  # type: ignore[misc]
                priority=incident.priority,
                component=incident.component,
                created_by=incident.created_by,
                commander=incident.created_by,
                incident=incident,
                event_ts=incident.created_at,
            )

            milestone_declared = IncidentUpdate(
                created_by=incident.created_by,
                incident=incident,
                event_type="declared",
                event_ts=incident.created_at,
            )
            first_incident_update.save()
            milestone_declared.save()

            incident.members.add(incident.created_by)

        # Either we have an incident or an error was thrown
        incident_created.send_robust(sender=__name__, incident=incident)
        return incident

    @staticmethod
    def search(
        queryset: QuerySet[Incident] | None, search_term: str
    ) -> tuple[QuerySet[Incident], bool]:
        """Search for incidents using a search term, on the title and description fields.

        Args:
            queryset (QuerySet[Incident] | None): Queryset to search in. If None, search in all incidents. The Queryset allows to search on a subset of incidents (already filtered).
            search_term (str): Search term.

        Returns:
            tuple[QuerySet[Incident], bool]: Queryset of incidents matching the search term, and a boolean indicating if the search may contain duplicates objects.
        """
        if queryset is None:
            queryset = Incident.objects.all()

        # If not search, return the original queryset
        if search_term is None or search_term.strip() == "":
            return queryset, False

        queryset_search_id = None

        # If the search is just an int, search for this ID + regular search
        try:
            search_term_as_int = int(search_term)
        except ValueError:
            pass
        else:
            queryset_search_id = deepcopy(queryset)
            queryset_search_id = queryset_search_id.filter(id=search_term_as_int)

        # Postgres search on title + description
        # XXX Improve search performance and relevance
        vector = SearchVector("title", config="english", weight="A") + SearchVector(
            "description", config="english", weight="B"
        )
        query = SearchQuery(search_term, config="english", search_type="websearch")
        queryset = (
            queryset.annotate(rank=SearchRank(vector, query))
            .filter(rank__gte=0.1)
            .order_by("-rank")
        )

        # Add the search by id to the search by text if needed
        if queryset_search_id:
            queryset |= queryset_search_id  # type: ignore[operator]

        return queryset, False


class Incident(models.Model):
    # pylint: disable=no-member
    objects: IncidentManager = IncidentManager()

    id = models.BigAutoField[int, int](primary_key=True, auto_created=True)
    title = models.CharField[str, str](max_length=128)
    description = models.TextField[str, str]()
    _status = models.IntegerField(
        db_column="status",
        choices=IncidentStatus.choices,
        default=IncidentStatus.OPEN,
        verbose_name="Status",
    )
    severity = models.ForeignKey(
        Severity,
        on_delete=models.PROTECT,
        help_text="Severity (legacy, superseded by priority)",
        null=True,
        blank=True,
    )
    severity.system_check_deprecated_details = {
        "msg": "The Incident.severity field has been deprecated.",
        "hint": "Use Incident.priority instead.",
        "id": "fields.W921",
    }
    priority = models.ForeignKey(
        Priority,
        on_delete=models.PROTECT,
        help_text="Priority",
    )
    component = models.ForeignKey(
        Component, on_delete=models.PROTECT
    )
    environment = models.ForeignKey(
        Environment, on_delete=models.PROTECT
    )
    created_by = models.ForeignKey(
        User,
        on_delete=models.PROTECT,
        related_name="incidents_created_by",
        blank=False,
    )

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    closed_at = models.DateTimeField(
        null=True, blank=True
    )  # XXX-ZOR make this an event

    # XXX-ZOR pick a more meaningful name. maybe 'hidden'
    # XXX-ZOR document intent and impl. status
    ignore = models.BooleanField(
        default=False,
        help_text="An ignored incident can be closed right away. In the future, it won't affect statistics or appear in reports. This is useful for duplicates incidents, or false-positives.",
    )
    private = models.BooleanField(
        default=False,
        help_text="A private incident is not communicated in #tech-incidents, and its created conversation is private. In the future, we may restrict the visibility to incident members only.",
    )
    tags = TaggableManager(blank=True)

    members = models.ManyToManyField["User", "IncidentMembership"](
        User,
        through=IncidentMembership,
        through_fields=("incident", "user"),
    )
    impacts = models.ManyToManyField["Impact", "IncidentImpact"](
        "Impact", through="IncidentImpact"
    )

    metrics = models.ManyToManyField["MetricType", "IncidentMetric"](
        MetricType,
        through="IncidentMetric",
    )
    roles = models.ManyToManyField[User, "IncidentRole"](
        User, through=IncidentRole, related_name="incident_roles_set"
    )

    if TYPE_CHECKING:
        roles_set: RelatedManager[IncidentRole]

    class Meta(TypedModelMeta):
        constraints = [
            models.CheckConstraint(
                name="%(app_label)s_%(class)s__status_valid",
                check=models.Q(_status__in=IncidentStatus.values),
            )
        ]

    def __str__(self) -> str:
        return f"#{self.id} - {self.title}"

    def get_absolute_url(self) -> str:
        return reverse("incidents:incident-detail", kwargs={"incident_id": self.id})

    @property
    def short_description(self) -> str:
        return Truncator(self.description).chars(100)

    @property
    def slack_channel_name(self) -> str | None:
        if hasattr(self, "conversation"):
            return self.conversation.name
        return None

    @property
    def canonical_name(self) -> str:
        if self.created_at is None:
            raise RuntimeError(
                "Incident must be saved before canonical_name can be computed"
            )
        return f"{localtime(self.created_at).strftime('%Y%m%d')}-{str(self.id)[:8]}"

    @property
    def status(self) -> IncidentStatus:
        return IncidentStatus(self._status)

    @status.setter
    def status(self, status: IncidentStatus) -> None:
        self._status = status

    @property
    def slack_channel_url(self) -> str | None:
        if hasattr(self, "conversation"):
            return self.conversation.link
        return None

    @property
    def status_page_url(self) -> str:
        """Similar with `get_absolute_url` but with full domain, to be used out of the website."""
        return f"{settings.BASE_URL}{self.get_absolute_url()}"

    @property
    def needs_postmortem(self) -> bool:
        return (
            bool(
                self.priority
                and self.environment
                and self.priority.needs_postmortem
                and self.environment.value == "PRD"
            )
            if apps.is_installed("firefighter.confluence")
            else False
        )

    @property
    def can_be_closed(self) -> tuple[bool, list[tuple[str, str]]]:
        # XXX-ZOR we should use a proper FSM abstraction
        cant_closed_reasons: list[tuple[str, str]] = []
        if self.ignore:
            return True, []
        if self.needs_postmortem:
            if self.status.value != IncidentStatus.POST_MORTEM:
                cant_closed_reasons.append(
                    (
                        "STATUS_NOT_POST_MORTEM",
                        f"Incident is not in PostMortem status, and needs one because of its priority and environment ({self.priority.name}/{self.environment.value}).",
                    )
                )
        elif self.status.value < IncidentStatus.FIXED:
            cant_closed_reasons.append(
                (
                    "STATUS_NOT_MITIGATED",
                    f"Incident is not in {IncidentStatus.FIXED.label} status (currently {self.status.label}).",
                )
            )
        missing_milestones = self.missing_milestones()
        if len(missing_milestones) > 0:
            cant_closed_reasons.append(
                (
                    "MISSING_REQUIRED_KEY_EVENTS",
                    f"Missing key events: {', '.join(missing_milestones)}",
                )
            )

        if len(cant_closed_reasons) > 0:
            return False, cant_closed_reasons
        return True, cant_closed_reasons

    @property
    def ask_for_milestones(self) -> bool:
        # XXX-ZOR we should use a proper FSM abstraction
        return len(self.missing_milestones()) > 0

    def missing_milestones(self) -> list[str]:
        """Returns all required Milestones still needed to compute the metrics."""
        if self.ignore:
            return []

        # We get a list of all event_type of IncidentUpdates...
        milestones_list = (
            IncidentUpdate.objects.filter(
                incident_id=self.id,
                event_ts__isnull=False,
                event_type__isnull=False,
            )
            .aggregate(milestone_list=ArrayAgg("event_type", distinct=True, default=[]))
            .get("milestone_list", [])
        )

        incident_milestones: set[str] = set(cast('list["str"]', milestones_list))
        required_milestone_types = set(
            MilestoneType.objects.filter(required=True).values_list(
                "event_type", flat=True
            )
        )
        # And we make the differences of the sets
        return list(required_milestone_types - incident_milestones)

    @property
    def latest_updates_by_type(self) -> dict[str, IncidentUpdate]:
        qs = (
            self.incidentupdate_set.order_by("event_type", "-event_ts")
            .distinct("event_type")
            .filter(event_type__isnull=False)
        )
        return {iu.event_type: iu for iu in qs}  # type: ignore

    @property
    def total_cost(self) -> int | float | Decimal:
        qs: QuerySet[IncidentCost] = self.incident_cost_set.exclude(amount__isnull=True)
        return sum(iu.amount for iu in qs if iu.amount is not None)

    def compute_metrics(self, *, purge: bool = False) -> None:
        """Compute all metrics (time to fix, ...) from events."""
        latest_updates_by_type = self.latest_updates_by_type

        for metric_type in MetricType.objects.all():
            lhs_type = metric_type.milestone_lhs.event_type
            rhs_type = metric_type.milestone_rhs.event_type

            lhs = latest_updates_by_type.get(lhs_type, None)
            rhs = latest_updates_by_type.get(rhs_type, None)

            if not lhs or not lhs.event_ts:
                logger.info(
                    f"Missing operand '{lhs_type}' on metric {metric_type.type} for #{self.id}"
                )
                if purge:
                    IncidentMetric.objects.filter(
                        incident=self, metric_type=metric_type
                    ).delete()
                continue
            if not rhs or not rhs.event_ts:
                logger.info(
                    f"Missing operand '{rhs_type}' on metric {metric_type.type} for #{self.id}"
                )
                if purge:
                    IncidentMetric.objects.filter(
                        incident=self, metric_type=metric_type
                    ).delete()
                continue
            duration = lhs.event_ts - rhs.event_ts
            if duration < TD0:
                logger.warning(
                    f"Tried to compute a negative metric! Metric {metric_type.type} for #{self.id} has a duration of {duration} ({lhs.event_type}:{lhs.event_ts} - {rhs.event_type}:{rhs.event_ts})"
                )
                if purge:
                    IncidentMetric.objects.filter(
                        incident=self, metric_type=metric_type
                    ).delete()
                continue
            IncidentMetric.objects.update_or_create(
                incident=self,
                metric_type=metric_type,
                defaults={"duration": duration},
            )
        self.save()

    def build_invite_list(self) -> list[User]:
        """Send a Django Signal to get the list of users to invite from different integrations (Slack, Confluence, PagerDuty...).

        Returns:
            list[User]: Potentially non-unique list of Users to invite
        """
        users_list: list[User] = []

        # Send signal to modules (Confluence, PagerDuty...)
        result_users: list[tuple[Any, Exception | Iterable[User]]] = (
            signals.get_invites.send_robust(sender=None, incident=self)
        )

        # Aggregate the results
        for provider in result_users:
            if isinstance(provider[1], BaseException):
                logger.warning(
                    f"Provider {provider[0]} returned an error getting Users to invite: {provider[1]}."
                )
                continue

            users_list.extend(provider[1])

        logger.debug(f"Get invites users list: {users_list}")

        return users_list

    def update_roles(
        self,
        updater: User,
        roles_mapping: dict[str, User | None] | dict[str, User],
    ) -> IncidentUpdate:
        """Update the roles related to an incident, and create an IncidentUpdate.
        For each role, provide a User or None.

        This function will update the incident, create an [IncidentUpdate][firefighter.incidents.models.incident_update.IncidentUpdate], and trigger the [incident_updated][firefighter.incidents.signals.incident_updated] signal, with `update_roles` sender.

        Args:
            updater (User): The user who is updating the roles.
            roles_mapping (dict[str, User | None], optional): A dict of roles to update, with the new User or None. Defaults to None.

        Returns:
            IncidentUpdate: The created IncidentUpdate with the updated roles.
        """
        updated_fields: list[str] = []

        # Handle roles
        for role_slug, assigned_user in roles_mapping.items():
            try:
                role_type = IncidentRoleType.objects.get(slug=role_slug)
            except IncidentRoleType.DoesNotExist:
                logger.warning(f"Unknown role type: {role_slug}")
                continue
            if assigned_user is not None:
                IncidentRole.objects.update_or_create(
                    incident=self,
                    role_type=role_type,
                    defaults={"user": assigned_user},
                )
                logger.debug(
                    f"Updated role {role_slug} to user ID={assigned_user.id} for #{self.id}"
                )
                updated_fields.append(f"{role_slug}_id")
            elif role_type.required:
                logger.warning(f"Cannot remove required role: {role_slug}")
            else:
                IncidentRole.objects.filter(incident=self, role_type=role_type).delete()
                updated_fields.append(f"{role_slug}_id")

        # Handle legacy roles with IncidentUpdate
        # XXX(dugab): Custom roles are not saved in IncidentUpdate at the moment

        incident_update = IncidentUpdate(
            incident=self,
            created_by=updater,
            commander=roles_mapping.get("commander"),
            communication_lead=roles_mapping.get("communication_lead"),
        )
        incident_update.save()

        incident_updated.send_robust(
            "update_roles",
            incident=self,
            incident_update=incident_update,
            updated_fields=updated_fields,
        )

        return incident_update

    def create_incident_update(
        self: Incident,
        message: str | None = None,
        status: int | None = None,
        priority_id: UUID | None = None,
        component_id: UUID | None = None,
        created_by: User | None = None,
        event_type: str | None = None,
        title: str | None = None,
        description: str | None = None,
        environment_id: str | None = None,
        event_ts: datetime | None = None,
    ) -> IncidentUpdate:
        updated_fields: list[str] = []

        def _update_incident_field(
            incident: Incident, field_name: str, value: Any, updated_fields: list[str]
        ) -> None:
            if value is not None:
                setattr(incident, field_name, value)
                updated_fields.append(field_name)

        _update_incident_field(self, "_status", status, updated_fields)
        _update_incident_field(self, "priority_id", priority_id, updated_fields)
        _update_incident_field(self, "component_id", component_id, updated_fields)
        _update_incident_field(self, "title", title, updated_fields)
        _update_incident_field(self, "description", description, updated_fields)
        _update_incident_field(self, "environment_id", environment_id, updated_fields)

        old_priority = self.priority if priority_id is not None else None

        if updated_fields:
            self.save(update_fields=[*updated_fields, "updated_at"])

        if not (updated_fields or message):
            raise ValueError("No updated fields or message provided.")
        with transaction.atomic():
            incident_update = IncidentUpdate(
                incident=self,
                status=status,  # type: ignore
                priority_id=priority_id,
                environment_id=environment_id,
                component_id=component_id,
                message=message,
                created_by=created_by,
                title=title,
                description=description,
                event_type=event_type,
                event_ts=event_ts,
            )
            incident_update.save()

            if status == IncidentStatus.FIXED:
                IncidentUpdate.objects.update_or_create(
                    incident_id=self.id,
                    event_type="recovered",
                    defaults={
                        "event_ts": incident_update.event_ts,
                        "created_by": incident_update.created_by,
                    },
                )
                self.compute_metrics()

        incident_updated.send_robust(
            "update_status",
            incident=self,
            incident_update=incident_update,
            updated_fields=updated_fields,
            old_priority=old_priority,
        )

        if self.status == IncidentStatus.CLOSED:
            logger.debug("Close incident: sending signal")
            incident_closed.send_robust(sender=__name__, incident=self)

        return incident_update

    if TYPE_CHECKING:
        if settings.ENABLE_CONFLUENCE:
            postmortem_for: PostMortem
            postmortem_for_id: UUID
        priority_id: UUID
        environment_id: UUID
        component_id: UUID
        conversation: IncidentChannel
        incidentupdate_set: QuerySet[IncidentUpdate]


def component_filter_choices_queryset(_: Any) -> QuerySet[Component]:
    """Queryset for choices of Components in IncidentFilterSet.
    Moved it as a function because models are not loaded when creating filters.
    """
    return (
        Component.objects.all()
        .select_related("group")
        .order_by(
            "group__order",
            "name",
        )
    )


class IncidentFilterSet(django_filters.FilterSet):
    """Set of filters for incidents, shared by Web UI and API."""

    id = django_filters.CharFilter(lookup_expr="iexact")

    status = MultipleChoiceFilter(
        choices=IncidentStatus.choices,
        label="Status",
        field_name="_status",
        widget=CustomCheckboxSelectMultiple,
        null_value="All Statuses",
    )
    environment = ModelMultipleChoiceFilter(
        queryset=Environment.objects.all(),
        widget=CustomCheckboxSelectMultiple,
    )
    priority = ModelMultipleChoiceFilter(
        queryset=Priority.objects.all(),
        widget=CustomCheckboxSelectMultiple,
    )
    group = ModelMultipleChoiceFilter(
        queryset=Group.objects.all(), field_name="component__group_id", label="Group"
    )
    component = ModelMultipleChoiceFilter(
        queryset=component_filter_choices_queryset,
        label="Issue category",
        widget=GroupedCheckboxSelectMultiple,
    )
    created_at = FFDateRangeSingleFilter(field_name="created_at")
    order_by = OrderingFilter(
        fields=["priority", "title", "id", "created_at", "created_by"]
    )

    # User filters are not shown in the UI, but are available in the API or query params for the Web UI
    created_by = ModelMultipleChoiceFilter(queryset=User.objects.all())

    search = django_filters.CharFilter(
        field_name="search", method="incident_search", label="Search"
    )

    @staticmethod
    def incident_search(
        queryset: QuerySet[Incident], _name: str, value: str
    ) -> QuerySet[Incident]:
        """Search incidents by title, description, and ID.

        Args:
            queryset (QuerySet[Incident]): Queryset to search in.
            _name:
            value (str): Value to search for.

        Returns:
            QuerySet[Incident]: Search results.
        """
        return Incident.objects.search(queryset=queryset, search_term=value)[0]

    class Meta(TypedModelMeta):
        model = Incident
        fields = {
            "created_at": ["gte", "lte", "lt", "gt", "exact"],
            "id": ["gte", "lte", "lt", "gt", "iexact", "in"],
            "_status": ["gte", "lte", "lt", "gt"],
            "environment__value": ["exact", "in"],
            "ignore": ["exact"],
        }



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Never

from django.dispatch.dispatcher import receiver
from django.utils import timezone

from firefighter.firefighter.utils import is_during_office_hours
from firefighter.slack.messages.slack_messages import SlackMessageIncidentDuringOffHours
from firefighter.slack.signals import incident_channel_done

if TYPE_CHECKING:
    from firefighter.slack.models.incident_channel import IncidentChannel

logger = logging.getLogger(__name__)


@receiver(signal=incident_channel_done)
def incident_channel_done_oncall(channel: IncidentChannel, **kwargs: Never) -> None:
    logger.debug("Signal incident_channel_done received by PagerDuty app.")

    if is_during_office_hours(timezone.localtime()):
        logger.debug("Skipping oncall message for now.")
        return
    channel.send_message_and_save(SlackMessageIncidentDuringOffHours())



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from firefighter.firefighter.utils import get_in
from firefighter.slack.slack_app import SlackApp
from firefighter.slack.views.modals import selectable_modals
from firefighter.slack.views.modals.base_modal.modal_utils import update_modal

if TYPE_CHECKING:
    from slack_bolt.context.ack.ack import Ack

    from firefighter.slack.views.modals.base_modal.base import SlackModal

app = SlackApp()

logger = logging.getLogger(__name__)


selectable_list: list[tuple[str, SlackModal]] = list(
    filter(
        lambda x: x[0] is not None,
        [
            (x.callback_id, x())
            for x in selectable_modals
            if isinstance(x.callback_id, str)
        ],
    )
)
selectable: dict[str, SlackModal] = dict(selectable_list)


@app.action("incident_update_select_incident")
def update_update_modal(ack: Ack, body: dict[str, Any]) -> None:
    """Reacts to the selection of an incident, in the select modal."""
    ack()
    logger.info(body)

    callback_id = get_in(body, "view.callback_id")
    if callback_id in selectable:
        view = selectable[callback_id].build_modal_with_context(
            body, callback_id=callback_id
        )
        update_modal(body=body, view=view)


@app.action("open_link")
def open_link(ack: Ack) -> None:
    """Does nothing. ack() is mandatory, even on buttons that open a URL."""
    ack()



# Generated by Django 4.2.11 on 2024-04-30 15:39

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("incidents", "0001_initial_oss"),
    ]

    operations = [
        migrations.AlterField(
            model_name="severity",
            name="name",
            field=models.CharField(max_length=128, unique=True),
        ),
        migrations.AlterField(
            model_name="user",
            name="password",
            field=models.CharField(max_length=128, verbose_name="password"),
        ),
        migrations.CreateModel(
            name="FeatureTeam",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                ("name", models.CharField(max_length=80)),
                ("jira_project_key", models.CharField(max_length=10, unique=True)),
            ],
            options={
                "unique_together": {("name", "jira_project_key")},
                "verbose_name": "Feature Team",
                "verbose_name_plural": "Feature Teams",
            },
        ),
    ]



from __future__ import annotations

import logging

from celery import shared_task

from firefighter.confluence.models import ConfluencePage
from firefighter.confluence.service import confluence_service
from firefighter.firefighter.utils import get_in

logger = logging.getLogger(__name__)


@shared_task(
    name="confluence.sync_pages_content", soft_time_limit=60 * 5, time_limit=60 * 7
)
def sync_pages_content() -> None:
    for runbook in ConfluencePage.objects.all():
        res = confluence_service.get_page_with_body_and_version(runbook.page_id)
        # Check if the page still exists, if not, delete it
        if res.get("statusCode", None) == 404 and get_in(res, ["data", "authorized"]):
            logger.info(f"Page {runbook.page_id} {runbook} not found, deleting it")
            runbook.delete()
            continue

        body = res.get("body", None)
        if body is None:
            logger.warning("No body found for page %s", runbook)
            continue
        runbook.body_export_view = get_in(res, ["body", "export_view", "value"])
        runbook.body_storage = get_in(res, ["body", "storage", "value"])
        runbook.body_view = get_in(res, ["body", "view", "value"])
        runbook.version = get_in(res, ["version"], {})
        runbook.save()



from __future__ import annotations

from django import forms

from firefighter.incidents.forms.utils import GroupedModelChoiceField
from firefighter.incidents.models import Component


class CloseIncidentForm(forms.Form):
    title = forms.CharField(
        label="Title",
        max_length=128,
        min_length=10,
        widget=forms.TextInput,
    )
    description = forms.CharField(
        label="Summary",
        widget=forms.Textarea,
        min_length=10,
        max_length=1200,
    )
    message = forms.CharField(
        label="Update message",
        widget=forms.Textarea,
        min_length=10,
        max_length=1200,
        required=False,
    )
    component = GroupedModelChoiceField(
        choices_groupby="group",
        label="Issue category",
        queryset=Component.objects.all()
        .select_related("group")
        .order_by(
            "group__order",
            "name",
        ),
    )



from __future__ import annotations

from django.apps import AppConfig


class ApiConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    label = "api"
    name = "firefighter.api"



from __future__ import annotations

import logging
from typing import cast

from celery import shared_task
from django.db.utils import IntegrityError

from firefighter.confluence.models import PostMortem
from firefighter.confluence.service import confluence_service
from firefighter.confluence.utils import (
    CONFLUENCE_PM_TITLE_REGEX,
    parse_postmortem_archive_title,
)
from firefighter.incidents.models.incident import Incident

logger = logging.getLogger(__name__)


@shared_task(name="confluence.sync_postmortems")
def sync_postmortems() -> None:
    all_pm = confluence_service.get_page_descendant_pages(
        confluence_service.POSTMORTEM_FOLDER_ID
    )
    pm_missing_incident = []
    pm_no_match = []

    for pm in all_pm:
        data = confluence_service.parse_confluence_page(pm)

        if "name" not in data or data["name"] is None or data["name"] == "":
            logger.error(f"PostMortem {pm['id']} has no name")
            continue

        # Ignore [ARCHIVE] folders
        if parse_postmortem_archive_title(data["name"]) is not None:
            continue

        # Ignore any template page
        if "[TEMPLATE]" in data["name"].upper():
            continue

        # Ignore any page whose title does not start with a #
        if not data["name"].startswith("#"):
            continue

        match = CONFLUENCE_PM_TITLE_REGEX.match(data["name"])
        if not match:
            pm_no_match.append(data["name"])
            continue

        match_dict = match.groupdict()
        incident_id = match_dict.get("id")
        if not incident_id:
            continue
        if not Incident.objects.filter(id=incident_id).exists():
            pm_missing_incident.append(data["name"])
            continue
        data_editable = cast("dict[str, str]", data)
        try:
            PostMortem.objects.update_or_create(
                page_id=int(data_editable.pop("page_id")),
                defaults=data_editable
                | {
                    "incident_id": incident_id,
                },
            )
        except IntegrityError:
            logger.exception(f"IntegrityError for {data} {pm}")
    logger.info(pm_missing_incident)
    logger.info(pm_no_match)



from __future__ import annotations

import datetime
import logging
from datetime import timedelta
from typing import TYPE_CHECKING, Any

from django.conf import settings
from django.dispatch import receiver
from django.utils import timezone
from slack_sdk.errors import SlackApiError

from firefighter.incidents.models.incident_membership import IncidentRole
from firefighter.incidents.models.incident_role_type import IncidentRoleType
from firefighter.incidents.signals import incident_updated
from firefighter.slack.messages.slack_messages import SlackMessageRoleAssignedToYou
from firefighter.slack.signals import incident_channel_done

if TYPE_CHECKING:
    from collections.abc import Iterable

    from firefighter.incidents.models.incident import Incident
    from firefighter.incidents.models.incident_update import IncidentUpdate
    from firefighter.incidents.models.user import User

logger = logging.getLogger(__name__)
TZ = timezone.get_current_timezone()

ROLE_REMINDER_MIN_DAYS_INTERVAL: int = settings.FF_ROLE_REMINDER_MIN_DAYS_INTERVAL


@receiver(signal=incident_channel_done)
def incident_created_roles_dm(sender: Any, incident: Incident, **kwargs: Any) -> None:
    # Send the SlackMessageRoleAssignedToYou to all users with roles, at incident creation

    if ROLE_REMINDER_MIN_DAYS_INTERVAL == -1:
        return

    for role in incident.roles_set.all():
        message = SlackMessageRoleAssignedToYou(
            incident=incident, role_type=role.role_type, first_update=True
        )
        if role.user.slack_user and not has_user_had_the_same_role_recently(
            role.user, role_type=role.role_type
        ):
            logger.debug(
                f"Sending message to {role.user} for role {role.role_type.name}"
            )
            try:
                role.user.slack_user.send_private_message(message)
            except SlackApiError:
                logger.exception(
                    f"Failed to send message to {role.user.id} for role {role.role_type.name}"
                )


@receiver(signal=incident_updated, sender="update_roles")
def incident_updated_roles_dm(
    sender: Any,
    incident: Incident,
    incident_update: IncidentUpdate,
    updated_fields: list[str],
    **kwargs: Any,
) -> None:
    updated_roles = _get_updated_roles(incident, updated_fields)
    for role in updated_roles:
        message = SlackMessageRoleAssignedToYou(
            incident=incident, role_type=role.role_type, first_update=False
        )
        if role.user.slack_user and not has_user_had_the_same_role_recently(
            role.user, role_type=role.role_type
        ):
            try:
                role.user.slack_user.send_private_message(message)
            except SlackApiError:
                logger.exception(
                    f"Failed to send message to {role.user.id} for role {role.role_type.name}"
                )


def _get_updated_roles(
    incident: Incident, updated_fields: list[str]
) -> Iterable[IncidentRole]:
    if len(updated_fields) == 0:
        return []

    incident_roles: list[IncidentRole] = []
    # XXX Smelly
    for incident_role_type in IncidentRoleType.objects.all().filter(
        slug__in=[updated_field.removesuffix("_id") for updated_field in updated_fields]
    ):
        # Get the IncidentRole if it exists in DB, or create it locally with User=None
        try:
            incident_role = IncidentRole.objects.select_related(
                "user", "user__slack_user"
            ).get(incident=incident, role_type=incident_role_type)
            incident_roles.append(incident_role)
        except IncidentRole.DoesNotExist:
            pass

    return incident_roles


def has_user_had_the_same_role_recently(
    user: User, role_type: IncidentRoleType
) -> bool:
    recently_td = timedelta(days=ROLE_REMINDER_MIN_DAYS_INTERVAL)
    before_recent = datetime.datetime.now(tz=TZ) - recently_td
    return IncidentRole.objects.filter(
        user=user, role_type=role_type, updated_at__gte=before_recent
    ).exists()



from __future__ import annotations

from firefighter.api.views._base import ReadOnlyModelViewSet
from firefighter.confluence.models import Runbook
from firefighter.confluence.serializers import RunbookSerializer


class RunbookViewSet(ReadOnlyModelViewSet[Runbook]):
    queryset = Runbook.objects.all()
    serializer_class = RunbookSerializer



from __future__ import annotations

import logging
from typing import Any

from firefighter.firefighter.utils import get_in
from firefighter.slack.models.conversation import Conversation, ConversationStatus
from firefighter.slack.slack_app import SlackApp

app = SlackApp()

logger = logging.getLogger(__name__)


@app.event("channel_archive")
@app.event("group_archive")
def channel_archive(event: dict[str, Any]) -> None:
    logger.debug(event)
    channel_id = get_in(event, "channel")
    if not channel_id:
        logger.warning(f"Invalid event! {event}")
        return

    Conversation.objects.filter(channel_id=channel_id).update(
        _status=ConversationStatus.ARCHIVED
    )



# ISC License

# Copyright (c) 2021, Timothée Mazzucotelli

# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.

# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

# Obtained and slightly modified from https://github.com/mkdocstrings/python/blob/main/scripts/gen_credits.py

"""Script to generate the project's credits."""

from __future__ import annotations

import operator
import os
import re
import tomllib
from importlib.metadata import PackageNotFoundError, metadata
from itertools import chain
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, cast

from jinja2 import StrictUndefined
from jinja2.sandbox import SandboxedEnvironment

if TYPE_CHECKING:
    from collections.abc import Mapping

project_dir = Path(os.getenv("MKDOCS_CONFIG_DIR", "."))
with project_dir.joinpath("pyproject.toml").open("rb") as pyproject_file:
    pyproject = tomllib.load(pyproject_file)
project = pyproject["project"]
pdm = pyproject["tool"]["pdm"]
with project_dir.joinpath("pdm.lock").open("rb") as lock_file:
    lock_data = tomllib.load(lock_file)
lock_pkgs = {pkg["name"].lower(): pkg for pkg in lock_data["package"]}
project_name = project["name"]
regex = re.compile(r"(?P<dist>[\w.-]+)(?P<spec>.*)$")


def _get_license(pkg_name: str) -> str:
    try:
        data = metadata(pkg_name)
    except PackageNotFoundError:
        return "?"
    license_name = cast("dict", data).get("License", "").strip()
    multiple_lines = bool(license_name.count("\n"))
    # TODO: Remove author logic once all my packages licenses are fixed.
    author = ""
    if multiple_lines or not license_name or license_name == "UNKNOWN":
        for header, value in cast("dict", data).items():
            if header == "Classifier" and value.startswith("License ::"):
                license_name = value.rsplit("::", 1)[1].strip()
            elif header == "Author-email":
                author = value
    if license_name == "Other/Proprietary License" and "pawamoy" in author:
        license_name = "ISC"
    return license_name or "?"


def _get_deps(base_deps: Mapping[str, Mapping[str, str]]) -> dict[str, dict[str, str]]:
    deps = {}
    for dep in base_deps:
        parsed = regex.match(dep).groupdict()  # type: ignore[union-attr]
        dep_name = parsed["dist"].lower()
        if dep_name not in lock_pkgs:
            continue
        deps[dep_name] = {
            "license": _get_license(dep_name),
            **parsed,
            **lock_pkgs[dep_name],
        }

    again = True
    while again:
        again = False
        for pkg_name, pkg_lock_pkgs in lock_pkgs.items():
            if pkg_name in deps:
                for pkg_dependency in pkg_lock_pkgs.get("dependencies", []):
                    parsed = regex.match(pkg_dependency).groupdict()  # type: ignore[union-attr]
                    dep_name = parsed["dist"].lower()
                    if (
                        dep_name in lock_pkgs
                        and dep_name not in deps
                        and dep_name != project["name"]
                    ):
                        deps[dep_name] = {
                            "license": _get_license(dep_name),
                            **parsed,
                            **pkg_lock_pkgs,
                        }
                        again = True

    return deps


def _render_credits() -> str:
    dev_dependencies = _get_deps(chain(*pdm.get("dev-dependencies", {}).values()))  # type: ignore[arg-type]
    prod_dependencies = _get_deps(
        chain(  # type: ignore[arg-type]
            project.get("dependencies", []),
            chain(*project.get("optional-dependencies", {}).values()),
        ),
    )

    template_data = {
        "project_name": project_name,
        "prod_dependencies": sorted(
            prod_dependencies.values(), key=operator.itemgetter("name")
        ),
        "dev_dependencies": sorted(
            dev_dependencies.values(), key=operator.itemgetter("name")
        ),
        "more_credits": None,
    }
    template_text = dedent(
        """
        # Credits

        These projects were used to build *{{ project_name }}*. **Thank you!**

        [`python`](https://www.python.org/) |
        [`pdm`](https://pdm.fming.dev/) |
        [`hatchling`](https://github.com/pypa/hatch)

        _Thanks to all the contributors to the JavaScript packages we use as well!_

        {% macro dep_line(dep) -%}
        [`{{ dep.name }}`](https://pypi.org/project/{{ dep.name }}/) | {{ dep.summary }} | {{ ("`" ~ dep.spec ~ "`") if dep.spec else "" }} | `{{ dep.version }}` | {{ dep.license }}
        {%- endmacro %}

        ### Runtime dependencies

        Project | Summary | Version (accepted) | Version (last resolved) | License
        ------- | ------- | ------------------ | ----------------------- | -------
        {% for dep in prod_dependencies -%}
        {{ dep_line(dep) }}
        {% endfor %}

        ### Development dependencies

        Project | Summary | Version (accepted) | Version (last resolved) | License
        ------- | ------- | ------------------ | ----------------------- | -------
        {% for dep in dev_dependencies -%}
        {{ dep_line(dep) }}
        {% endfor %}

        {% if more_credits %}**[More credits from the author]({{ more_credits }})**{% endif %}
        """,
    )
    jinja_env = SandboxedEnvironment(undefined=StrictUndefined)
    return jinja_env.from_string(template_text).render(**template_data)


print(_render_credits())  # noqa: T201



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django.contrib import messages
from rest_framework.authtoken.admin import TokenAdmin
from rest_framework.authtoken.models import TokenProxy

from firefighter.api.models import APITokenProxy
from firefighter.firefighter.admin import admin_custom as admin
from firefighter.incidents.models.user import User

logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from django.db.models import ForeignKey
    from django.db.models.query import QuerySet
    from django.forms import ModelChoiceField, ModelForm
    from django.http.request import HttpRequest as BaseHttpRequest

    class HttpRequest(BaseHttpRequest):
        user: User


class APITokenAdmin(TokenAdmin):
    """Custom Admin for DRF Token.
    Add supports for custom permissions.
    """

    def formfield_for_foreignkey(
        self,
        db_field: ForeignKey[Any, Any],
        request: HttpRequest,  # type: ignore[override]
        **kwargs: Any,
    ) -> ModelChoiceField:  # type: ignore[type-arg]
        """Show all or only current user depending on permissions."""
        if db_field.name == "user":
            if request.user.has_perm("api.can_add_any") or request.user.has_perm(
                "api.can_edit_any"
            ):
                kwargs["queryset"] = User.objects.all()
            elif request.user.has_perm("api.can_add_own"):
                kwargs["queryset"] = User.objects.filter(id=request.user.id)
        return super().formfield_for_foreignkey(db_field, request, **kwargs)  # type: ignore[return-value]

    def get_form(
        self,
        request: HttpRequest,  # type: ignore[override]
        obj: APITokenProxy | None = None,
        change: bool = False,  # noqa: FBT001, FBT002
        **kwargs: Any,
    ) -> type[ModelForm[APITokenProxy]]:
        """Prefill the form with the current user."""
        form: type[ModelForm[APITokenProxy]] = super().get_form(
            request, obj, change, **kwargs
        )
        form.base_fields["user"].initial = request.user
        return form

    def get_queryset(self, request: HttpRequest) -> QuerySet[APITokenProxy]:  # type: ignore[override]
        """Show all or only own tokens depending on permissions."""
        qs = super().get_queryset(request)
        if request.user.has_perm("api.can_view_any"):
            return qs
        return qs.filter(user=request.user)

    def get_sortable_by(self, request: HttpRequest):  # type: ignore[no-untyped-def,override]
        """Hack to send a message depending on the status of the user."""
        if request.user.has_perm("api.can_view_any"):
            self.message_user(request, "You are seeing all tokens.", messages.WARNING)
        elif request.user.has_perm("api.can_view_own"):
            self.message_user(
                request, "You are only seeing your tokens.", messages.WARNING
            )
        return super().get_sortable_by(request)

    def has_view_permission(
        self,
        request: HttpRequest,  # type: ignore[override]
        obj: APITokenProxy | None = None,
    ) -> bool:
        if obj is None:
            return request.user.has_perm("api.can_view_any") or request.user.has_perm(
                "api.can_view_own"
            )
        if request.user.has_perm("api.can_view_any"):
            return True
        if request.user.has_perm("api.can_view_own"):
            return bool(obj.user == request.user)
        return False

    def has_add_permission(self, request: HttpRequest) -> bool:  # type: ignore[override]
        return request.user.has_perm("api.can_add_any") or request.user.has_perm(
            "api.can_add_own"
        )

    def has_delete_permission(
        self,
        request: HttpRequest,  # type: ignore[override]
        obj: APITokenProxy | None = None,
    ) -> bool:
        if obj is None:
            return request.user.has_perm("api.can_delete_any") or request.user.has_perm(
                "api.can_delete_own"
            )
        if request.user.has_perm("api.can_delete_any"):
            return True
        if request.user.has_perm("api.can_delete_own"):
            return bool(obj.user == request.user)
        return False

    def has_change_permission(
        self,
        request: HttpRequest,  # type: ignore[override]
        _obj: APITokenProxy | None = None,
    ) -> bool:
        return request.user.has_perm("api.can_edit_any")


# Remove the default TokenAdmin created by DRF and replace it with our custom one.
admin.site.unregister(TokenProxy)
admin.site.register(APITokenProxy, APITokenAdmin)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from django.conf import settings
from slack_sdk.models.blocks.basic_components import MarkdownTextObject
from slack_sdk.models.blocks.block_elements import ButtonElement
from slack_sdk.models.blocks.blocks import (
    ActionsBlock,
    Block,
    ContextBlock,
    DividerBlock,
    SectionBlock,
)
from slack_sdk.models.views import View

from firefighter.incidents.enums import IncidentStatus
from firefighter.incidents.forms.close_incident import CloseIncidentForm
from firefighter.slack.slack_templating import slack_block_footer, slack_block_separator
from firefighter.slack.utils import respond
from firefighter.slack.views.modals.base_modal.base import ModalForm
from firefighter.slack.views.modals.base_modal.mixins import (
    IncidentSelectableModalMixin,
)
from firefighter.slack.views.modals.postmortem import PostMortemModal
from firefighter.slack.views.modals.update_status import UpdateStatusModal

if TYPE_CHECKING:
    from slack_bolt.context.ack.ack import Ack

    from firefighter.incidents.models.incident import Incident, User
    from firefighter.slack.views.modals.base_modal.form_utils import (
        SlackFormAttributesDict,
    )

logger = logging.getLogger(__name__)


class CloseIncidentFormSlack(CloseIncidentForm):
    slack_fields: SlackFormAttributesDict = {
        "message": {
            "input": {
                "multiline": True,
                "placeholder": "Please describe the new status for the incident.\nE.g: Fixed with instance reboot.",
            },
            "block": {"hint": None},
        },
        "description": {
            "input": {
                "multiline": True,
                "placeholder": "Help people responding to the incident. This will be posted to #tech-incidents and on our internal status page.\nThis description can be edited later.",
            },
            "block": {"hint": None},
        },
        "environment": {
            "input": {
                "placeholder": "Select an environment",
            },
            "widget": {
                "label_from_instance": lambda obj: f"{obj.value} - {obj.description}",
            },
        },
    }


class CloseModal(
    IncidentSelectableModalMixin,
    ModalForm[CloseIncidentFormSlack],
):
    open_action: str = "close_incident"
    open_shortcut = "close_incident"
    callback_id: str = "incident_close"

    form_class = CloseIncidentFormSlack

    def build_modal_fn(
        self, body: dict[str, Any], incident: Incident, **kwargs: Any
    ) -> View:
        can_be_closed, reasons = incident.can_be_closed
        if not can_be_closed:
            reason_blocks: list[Block] = []
            for reason in reasons:
                if reason[0] == "MISSING_REQUIRED_KEY_EVENTS":
                    reason_blocks += [
                        SectionBlock(
                            text=":warning: *Missing required key events* :warning:"
                        ),
                        ContextBlock(
                            elements=[
                                MarkdownTextObject(
                                    text=reason[1]
                                    + "\nPlease fill the required key events of the incident before closing it."
                                )
                            ]
                        ),
                    ]
                elif reason[0] == "STATUS_NOT_POST_MORTEM":
                    reason_blocks += [
                        SectionBlock(
                            text=f":warning: *Status is not _{IncidentStatus.POST_MORTEM.label}_* :warning:\n"
                        ),
                        ContextBlock(
                            elements=[
                                MarkdownTextObject(
                                    text=f"This status is required. Please update the status to _{IncidentStatus.POST_MORTEM.label}_ and *fill the post-mortem on Confluence* before closing the incident."
                                )
                            ]
                        ),
                        ActionsBlock(
                            elements=[
                                (
                                    ButtonElement(
                                        text="Fill the post-mortem on Confluence",
                                        action_id="open_link",
                                        url=incident.postmortem_for.page_edit_url,
                                        style="primary",
                                    )
                                    if hasattr(incident, "postmortem_for")
                                    else ButtonElement(
                                        text="Create the post-mortem",
                                        action_id=str(PostMortemModal.push_action),
                                        value=str(incident.id),
                                        style="primary",
                                    )
                                ),
                                ButtonElement(
                                    text="Update status",
                                    action_id=UpdateStatusModal.push_action,
                                    value=str(incident.id),
                                ),
                            ]
                        ),
                        ContextBlock(
                            elements=[
                                MarkdownTextObject(
                                    text=f":bulb: You can also publish an update and change the status of the incident with `{settings.SLACK_INCIDENT_COMMAND} update`. You can also use `{settings.SLACK_INCIDENT_COMMAND} postmortem` to access or create your post-mortem."
                                )
                            ]
                        ),
                    ]
                elif reason[0] == "STATUS_NOT_MITIGATED":
                    reason_blocks += [
                        SectionBlock(
                            text=f":warning: *Status is not _{IncidentStatus.FIXED.label}_* :warning:\n"
                        ),
                        ContextBlock(
                            elements=[
                                MarkdownTextObject(
                                    text=f"You can only close an incident when its status is _{IncidentStatus.FIXED.label}_ or _{IncidentStatus.POST_MORTEM.label}_. The _{IncidentStatus.POST_MORTEM.label}_ status is not mandatory for this incident."
                                )
                            ]
                        ),
                        ActionsBlock(
                            elements=[
                                ButtonElement(
                                    text="Update status",
                                    action_id=UpdateStatusModal.push_action,
                                    value=str(incident.id),
                                    style="primary",
                                ),
                            ]
                        ),
                        ContextBlock(
                            elements=[
                                MarkdownTextObject(
                                    text=f":bulb: You can also publish an update and change the status of the incident with `{settings.SLACK_INCIDENT_COMMAND} update`"
                                )
                            ]
                        ),
                    ]

            return View(
                type="modal",
                title=f"Close incident #{incident.id}"[:24],
                submit=None,
                callback_id=self.callback_id,
                private_metadata=str(incident.id),
                blocks=[
                    SectionBlock(
                        text=":x:  This incident can't be closed yet.\n:point_right:  Please fix the following issues before closing it.",
                    ),
                    DividerBlock(),
                    *reason_blocks,
                    slack_block_separator(),
                    slack_block_footer(),
                ],
            )
        logger.debug(body)
        context_elements = [
            MarkdownTextObject(
                text=":warning: You're about to close this incident which will archive the Slack channel and execute any post-incident actions.\nPlease review the incident details below."
            ),
        ]
        if hasattr(incident, "postmortem_for"):
            context_elements.append(
                MarkdownTextObject(
                    text=f":page_facing_up: Make sure <{incident.postmortem_for.page_url}|your post-mortem> is complete before closing the incident."
                )
            )
        blocks: list[Block] = [
            ContextBlock(elements=context_elements),
            *self.get_form_class()(
                initial=self._get_initial_form_values(incident)
            ).slack_blocks(),
            slack_block_separator(),
            slack_block_footer(),
        ]

        return View(
            type="modal",
            title=f"Close incident #{incident.id}"[:24],
            submit="Close incident"[:24],
            callback_id=self.callback_id,
            private_metadata=str(incident.id),
            blocks=blocks,
        )

    def handle_modal_fn(  # type: ignore[override]
        self, ack: Ack, body: dict[str, Any], incident: Incident, user: User
    ) ->  bool | None:
        """Handle response from /incident close modal."""
        slack_form = self.handle_form_errors(
            ack, body, forms_kwargs={"initial": self._get_initial_form_values(incident)}
        )
        if slack_form is None:
            return None
        form = slack_form.form
        # If fields haven't changed, don't include them in the update.
        update_kwargs = {}
        for changed_key in form.changed_data:
            if changed_key == "component":
                update_kwargs["component_id"] = form.cleaned_data[changed_key].id
            if changed_key in {"description", "title", "message"}:
                update_kwargs[changed_key] = form.cleaned_data[changed_key]
        # Check can close
        can_close, reasons = incident.can_be_closed
        if not can_close:
            logger.warning(
                f"Tried to close an incident that can't be closed yet! Aborting. Incident #{incident.id}. Reasons: {reasons}"
            )
            respond(
                body=body,
                text=f"It looks like this incident #{incident.id} could not be closed.\nReasons: {reasons}.\nPlease tell @pulse (#tech-pe-pulse) if you think this is an error.",
            )
            return False
        self._trigger_incident_workflow(incident, user, update_kwargs)
        return None

    def get_select_modal_title(self) -> str:
        return "Close incident"

    def get_select_title(self) -> str:
        return "Select a critical incident to close"

    @staticmethod
    def _get_initial_form_values(incident: Incident) -> dict[str, Any]:
        return {
            "title": incident.title,
            "description": incident.description,
            "component": incident.component,
        }

    @staticmethod
    def _trigger_incident_workflow(
        incident: Incident, user: User, update_kwargs: dict[str, Any]
    ) -> None:
        incident.create_incident_update(
            created_by=user,
            status=IncidentStatus.CLOSED,
            **update_kwargs,
        )


modal_close = CloseModal()



from __future__ import annotations

from django.urls import path

from firefighter.pagerduty.views.oncall_list import OncallListView
from firefighter.pagerduty.views.oncall_trigger import CreatePagerDutyIncidentFreeView

app_name = "pagerduty"
urlpatterns = [
    path("oncall/", OncallListView.as_view(), name="oncall-list"),
    path(
        "oncall/trigger/",
        CreatePagerDutyIncidentFreeView.as_view(),
        name="oncall_trigger",
    ),
]



from __future__ import annotations

import logging
import uuid
from itertools import groupby
from typing import Any, ClassVar

from django.db import models
from django.db.utils import IntegrityError
from django.utils import timezone
from django_stubs_ext.db.models import TypedModelMeta

from firefighter.incidents.models.incident import Incident
from firefighter.incidents.models.user import User

logger = logging.getLogger(__name__)

STATUS_HELP_TEXT = """The current state of the Service. Valid statuses are:
<ul>
<li><code>active</code>: The service is enabled and has no open incidents. This is the only status a service can be created with.</li>
<li><code>warning</code>: The service is enabled and has one or more acknowledged incidents.</li>
<li><code>critical</code>: The service is enabled and has one or more triggered incidents.</li>
<li><code>maintenance</code>: The service is under maintenance, no new incidents will be triggered during maintenance mode.</li>
<li><code>disabled</code>: The service is disabled and will not have any new triggered incidents.</li>
</ul>"""


class PagerDutyService(models.Model):
    """A Service represents an entity you monitor (such as a web Service, email Service, or database Service.)
    It is a container for related Incidents that associates them with Escalation Policies.
    A Service is the focal point for Incident management; Services specify the configuration for the behavior of Incidents triggered on them.
    This behavior includes specifying urgency and performing automated actions based on time of day, Incident duration, and other factors.

    - [Read more about Services in the PagerDuty Knowledge Base.](https://support.pagerduty.com/hc/en-us/sections/200550800-Services)
    """

    objects: ClassVar[models.Manager[PagerDutyService]]

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=128)
    status = models.CharField(max_length=16, help_text=STATUS_HELP_TEXT)  # XXX enum
    summary = models.CharField(
        max_length=256,
        help_text="A short-form, PD server-generated string that provides succinct, important information about an object suitable for primary labeling of an entity in a client. In many cases, this will be identical to name, though it is not intended to be an identifier. ",
    )
    api_url = models.URLField(
        max_length=256,
        help_text="The API show URL at which the object is accessible. Corresponds to PagerDuty API field `self`",
    )
    web_url = models.URLField(
        max_length=256,
        help_text="A URL at which the entity is uniquely displayed in the Web app. Corresponds to PagerDuty API field `html_url`",
    )
    pagerduty_id = models.CharField(
        max_length=64,
        help_text="PagerDuty ID for the service.",
        unique=True,
        db_index=True,
    )
    # TODO Escalation policy should not be null after migrations
    escalation_policy = models.ForeignKey(
        "PagerDutyEscalationPolicy", on_delete=models.CASCADE, null=True
    )

    ignore = models.BooleanField(
        default=False,
        help_text="Ignore this service. Ignored services can't be triggered, and are hidden ins most places.",
    )
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta(TypedModelMeta):
        verbose_name = "PagerDuty service"
        verbose_name_plural = "PagerDuty services"

    def __str__(self) -> str:
        return f"{self.summary} ({self.pagerduty_id})"

    def get_absolute_url(self) -> str:
        return self.web_url


class PagerDutyIncident(models.Model):
    """An Incident represents a problem or an issue that needs to be addressed and resolved.

    Incidents can be thought of as a problem or an issue within your Service that needs to be addressed and resolved, they are normalized and de-duplicated.

    Incidents can be triggered, acknowledged, or resolved, and are assigned to a User based on the Service's Escalation Policy.

    A triggered Incident prompts a Notification to be sent to the current On-Call User(s) as defined in the Escalation Policy used by the Service.

    Incidents are triggered through the Events API or are created by Integrations.

    - [Read more about Incidents in the PagerDuty Knowledge Base.](https://support.pagerduty.com/hc/en-us/articles/202829250-What-Is-an-Incident-)
    """

    objects: ClassVar[models.Manager[PagerDutyIncident]]
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    title = models.CharField(
        max_length=256,
        help_text="A succinct description of the nature, symptoms, cause, or effect of the incident.",
    )
    summary = models.CharField(
        max_length=256,
        help_text="A short-form, server-generated string that provides succinct, important information about an object suitable for primary labeling of an entity in a client. In many cases, this will be identical to name, though it is not intended to be an identifier.",
    )
    service = models.ForeignKey(PagerDutyService, on_delete=models.PROTECT)
    status = models.CharField(
        max_length=16,
        help_text="The current status of the incident. Allowed values: triggered, acknowledged, resolved",
    )  # XXX enum
    urgency = models.CharField(
        max_length=8,
        help_text="The urgency of the incident. Allowed values: high, low",
    )
    details = models.CharField(
        max_length=4000,
        help_text="Additional incident details. Corresponds to PagerDuty API field `body.details`",
    )
    api_url = models.URLField(
        max_length=256,
        help_text="The API show URL at which the object is accessible. Corresponds to PagerDuty API field `self`",
    )
    web_url = models.URLField(
        max_length=256,
        help_text="A URL at which the entity is uniquely displayed in the Web app. Corresponds to PagerDuty API field `html_url`",
    )
    incident_key = models.CharField(
        max_length=128,
        help_text="A string which identifies the incident. Sending subsequent requests referencing the same service and with the same incident_key will result in those requests being rejected if an open incident matches that incident_key.",
        unique=True,
    )
    incident_number = models.IntegerField(
        help_text="The number of the incident. This is unique across your account."
    )
    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    incident = models.ForeignKey(
        Incident,
        on_delete=models.CASCADE,
        related_name="pagerduty_incident_set",
        null=True,
    )

    class Meta(TypedModelMeta):
        verbose_name = "PagerDuty incident"
        verbose_name_plural = "PagerDuty incidents"

    def __str__(self) -> str:
        return f"{self.summary} ({self.incident_key})"

    def get_absolute_url(self) -> str:
        return self.web_url


class PagerDutyUserManager(models.Manager["PagerDutyUser"]):
    def get_or_none(self, **kwargs: Any) -> PagerDutyUser | None:
        try:
            return self.get(**kwargs)
        except PagerDutyUser.DoesNotExist:
            return None

    @staticmethod
    def upsert_by_pagerduty_id(
        pagerduty_id: str,
        email: str,
        phone_number: str,
        name: str,
    ) -> User | None:
        """Returns a User from PagerDuty info.
        It will update a PagerDutyUser and its associated User if the name, phone, PagerDuty team changes.
        """
        # Get a user by its email and update the name. Create the user if necessary.
        ff_user, _ = User.objects.update_or_create(email=email, defaults={"name": name})

        # Update or create a PD User, with the key being its user. Update other fields.
        try:
            pd_user, _ = PagerDutyUser.objects.update_or_create(
                user=ff_user,
                defaults={
                    "pagerduty_id": pagerduty_id,
                    "phone_number": phone_number,
                },
            )

            if pd_user.user == ff_user:
                return ff_user
            logger.warning(
                "PD and FF users not matching. PDID=%s, email=%s",
                pagerduty_id,
                email,
            )

        except IntegrityError:
            logger.warning(
                "IntegrityError! Could not upsert PagerDuty User. PDID=%s, email=%s.",
                pagerduty_id,
                email,
                exc_info=True,
            )

            pd_user, _ = PagerDutyUser.objects.update_or_create(
                pagerduty_id=pagerduty_id,
                defaults={
                    "user": ff_user,
                    "phone_number": phone_number,
                },
            )
            return ff_user
        return None

    @staticmethod
    def get_current_on_call_users_l1() -> list[User]:
        """Returns the list of oncall first responders in each escalation policy. Only the lowest escalation level user is returned per Escalation Policy.

        These users have their PagerDutyUser associated.
        """
        ep_user = PagerDutyOncall.objects.get_current_oncalls_per_escalation_policy()
        users: list[User] = [oncall[1][0].pagerduty_user.user for oncall in ep_user]

        return users


class PagerDutyTeam(models.Model):

    objects: ClassVar[models.Manager[PagerDutyTeam]]
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=128, unique=True)
    description = models.CharField(max_length=1024, null=True, blank=True)
    pagerduty_id = models.CharField(max_length=128, unique=True, db_index=True)
    pagerduty_api_url = models.URLField(max_length=256, null=True, blank=True)
    pagerduty_url = models.URLField(max_length=256, null=True, blank=True)

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta(TypedModelMeta):
        verbose_name = "PagerDuty team"
        verbose_name_plural = "PagerDuty teams"

    def __str__(self) -> str:
        return self.name


class PagerDutyUser(models.Model):
    objects: PagerDutyUserManager = PagerDutyUserManager()

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    user = models.OneToOneField(
        User, on_delete=models.CASCADE, related_name="pagerduty_user"
    )
    name = models.CharField(max_length=256, null=True, blank=True)
    pagerduty_id = models.CharField(max_length=128, unique=True, db_index=True)
    pagerduty_api_url = models.URLField(max_length=256, null=True, blank=True)
    pagerduty_url = models.URLField(max_length=256, null=True, blank=True)
    phone_number = models.CharField(max_length=128, blank=True, default="")
    teams = models.ManyToManyField[PagerDutyTeam, PagerDutyTeam](
        PagerDutyTeam, related_name="pagerduty_user_set", blank=True
    )

    class Meta(TypedModelMeta):
        verbose_name = "PagerDuty user"
        verbose_name_plural = "PagerDuty users"

    def __str__(self) -> str:
        return self.pagerduty_id


class PagerDutyEscalationPolicy(models.Model):
    """An Escalation Policy determines what User or Schedule will be Notified and in what order. This will happen when an Incident is triggered.

    Escalation Policies can be used by one or more Services.

    *Escalation Rules*

    - An Escalation Policy is made up of multiple Escalation Rules. Each Escalation Rule represents a level of On-Call duty.
    - It specifies one or more Users or Schedules to be notified when an unacknowledged Incident reaches that Escalation Rule.
    - The first Escalation Rule in the Escalation Policy is the User that will be notified about the triggered Incident.
    - If all On-Call User for a given Escalation Rule have been acknowledged of an Incident and the Escalation Rule's escalation delay has elapsed, the Incident escalates to the next Escalation Rule.

    - [Read more about Escalation Policies in the PagerDuty Knowledge Base](https://support.pagerduty.com/hc/en-us/articles/202828950-What-is-an-escalation-policy-)
    - [API Schema](https://developer.pagerduty.com/api-reference/c2NoOjI3NDgwMjE-escalation-policy)
    """

    objects: ClassVar[models.Manager[PagerDutyEscalationPolicy]]
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    pagerduty_id = models.CharField(max_length=128, unique=True, db_index=True)

    name = models.CharField(max_length=128, unique=True)
    summary = models.CharField(max_length=128, null=True, blank=True)
    description = models.CharField(max_length=1024, null=True, blank=True)

    pagerduty_api_url = models.URLField(max_length=256, null=True, blank=True)
    pagerduty_url = models.URLField(max_length=256, null=True, blank=True)

    teams = models.ManyToManyField(
        PagerDutyTeam, related_name="pagerduty_escalation_policy_set", blank=True
    )

    class Meta(TypedModelMeta):
        verbose_name_plural = "PagerDuty escalation policies"
        verbose_name = "PagerDuty escalation policy"

    def __str__(self) -> str:
        return f"{self.name} ({self.pagerduty_id})"


class PagerDutySchedule(models.Model):
    """A Schedule determines the time periods that Users are On-Call.

    Only On-Call Users are eligible to receive Notifications from firefighter.incidents.

    The details of the On-Call Schedule specify which single User is On-Call for that Schedule at any given point in time.

    An On-Call Schedule consists of one or more Schedule Layers that rotate a group of Users through the same shift at a set interval.

    Schedules are used by Escalation Policies as an escalation target for a given Escalation Rule.

    - [Read more about On-Call Schedules in the PagerDuty Knowledge Base](https://support.pagerduty.com/hc/en-us/sections/200550790-On-Call-Schedules)
    - [API Schema](https://developer.pagerduty.com/api-reference/c2NoOjI3NDgwMzU-schedule)
    """

    objects: ClassVar[models.Manager[PagerDutySchedule]]

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    pagerduty_id = models.CharField(max_length=128, unique=True, db_index=True)

    summary = models.CharField(max_length=128, null=True, blank=True)

    pagerduty_api_url = models.URLField(max_length=256, null=True, blank=True)
    pagerduty_url = models.URLField(max_length=256, null=True, blank=True)
    # TODO: Add more fields

    escalation_policies = models.ManyToManyField(
        PagerDutyEscalationPolicy, related_name="pagerduty_schedule_set", blank=True
    )
    users = models.ManyToManyField(
        PagerDutyUser, related_name="pagerduty_schedule_set", blank=True
    )
    teams = models.ManyToManyField(
        PagerDutyTeam, related_name="pagerduty_schedule_set", blank=True
    )

    class Meta(TypedModelMeta):
        verbose_name = "PagerDuty schedule"
        verbose_name_plural = "PagerDuty schedules"

    def __str__(self) -> str:
        return f"{self.summary} ({self.pagerduty_id})"


class PagerDutyOncallManager(models.Manager["PagerDutyOncall"]):
    def get_current_oncalls_per_escalation_policy(
        self,
    ) -> list[tuple[PagerDutyEscalationPolicy, list[PagerDutyOncall]]]:
        """Returns the list of on call Users. These users have their PagerDutyUser associated."""
        oncalls = (
            self.select_related(
                "pagerduty_user",
                "schedule",
                "escalation_policy",
                "pagerduty_user",
                "pagerduty_user__user__slack_user",
            )
            .filter(
                models.Q(start__lte=timezone.now(), end__gte=timezone.now())
                | models.Q(end__isnull=True)
            )
            .exclude(escalation_policy__pagerdutyservice__ignore__exact=True)
            .order_by("escalation_level")
            .order_by(
                "escalation_policy__pagerdutyservice",
                "pagerduty_user",
            )
            .distinct("escalation_policy__pagerdutyservice", "pagerduty_user")
        )
        oncalls_grouped: list[
            tuple[PagerDutyEscalationPolicy, list[PagerDutyOncall]]
        ] = []
        for escalation_policy, oncalls_grouper in groupby(
            oncalls, key=lambda x: x.escalation_policy
        ):
            oncalls_list = sorted(oncalls_grouper, key=lambda x: x.escalation_level)
            oncalls_grouped.append((escalation_policy, oncalls_list))
        return oncalls_grouped

    def get_current_oncalls_per_escalation_policy_name_first_responder(
        self,
    ) -> dict[str, User]:
        res = self.get_current_oncalls_per_escalation_policy()
        # Get only the first
        return {
            ep.name: oncalls[0].pagerduty_user.user
            for ep, oncalls in res
            if len(oncalls) > 0
        }


class PagerDutyOncall(models.Model):
    """An On-Call represents a contiguous unit of time for which a User will be On-Call for a given Escalation Policy and Escalation Rule.

    This may be the result of that User always being On-Call for the Escalation Rule, or a block of time during which the computed result of a Schedule on that Escalation Rule puts the User On-Call.

    During an On-Call, the User is expected to bear responsibility for responding to any Notifications they receive and working to resolve the associated Incident(s).

    On-Calls cannot be created directly through the API; they are the computed result of how Escalation Policies and Schedules are configured. The API provides read-only access to the On-Calls generated by PagerDuty.

    - [Read more about On-Call Schedules in the PagerDuty Knowledge Base](https://support.pagerduty.com/hc/en-us/sections/200550790-On-Call-Schedules)
    - [API Schema](https://developer.pagerduty.com/api-reference/c2NoOjI3NDgwNDc-oncall)
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    updated_at = models.DateTimeField(auto_now=True)
    created_at = models.DateTimeField(auto_now_add=True)

    escalation_policy = models.ForeignKey(
        PagerDutyEscalationPolicy, on_delete=models.CASCADE, related_name="oncall_set"
    )
    pagerduty_user = models.ForeignKey(
        PagerDutyUser, on_delete=models.CASCADE, related_name="oncall_set"
    )
    schedule = models.ForeignKey(
        PagerDutySchedule,
        on_delete=models.CASCADE,
        related_name="oncall_set",
        null=True,
        blank=True,
    )
    escalation_level = models.IntegerField(
        help_text="The escalation level for the on-call."
    )
    start = models.DateTimeField(
        null=True,
        blank=True,
        help_text="The start of the on-call. If null, the on-call is a permanent user on-call.",
    )
    end = models.DateTimeField(
        null=True,
        blank=True,
        help_text="The end of the on-call. If null, the user does not go off-call.",
    )

    objects = PagerDutyOncallManager()

    class Meta(TypedModelMeta):
        verbose_name = "PagerDuty on-call"
        verbose_name_plural = "PagerDuty on-calls"

    def __str__(self) -> str:
        return f"{self.pagerduty_user} ({self.escalation_policy}) - {self.start} to {self.end}"

    @property
    def user_name(self) -> str | None:
        return self.pagerduty_user.name
